﻿
Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D
Imports System.Threading


Module Bullet

    ' ===============================================================================================
    '   INIT / DESTROY / STEP
    ' ===============================================================================================

    Private Declare Sub Bullet_Init Lib "BulletWrapper.dll" (ByVal worldMin As Vector3, _
                                                             ByVal worldMax As Vector3, _
                                                             ByVal Gravity As Vector3)

    Friend Declare Sub Bullet_Destroy Lib "BulletWrapper.dll" ()

    Friend Declare Sub Bullet_Step Lib "BulletWrapper.dll" (ByVal elapsedTime As Single, _
                                                            ByVal maxSubSteps As Int32, _
                                                            Optional ByVal fixedTimestep As Single = 1 / 60)

    Friend Declare Function Bullet_GetNumObjects Lib "BulletWrapper.dll" () As Int32


    ' ===============================================================================================
    '   BASE PLANE
    ' ===============================================================================================

    Friend Declare Function Bullet_CreateFlatPlane Lib "BulletWrapper.dll" (ByVal PlaneHeight As Single) As Int32


    ' ===============================================================================================
    '   SIMPLE-SHAPES ( BOX / SPHERE / CAPSULE / CONE / CYLINDER )
    ' ===============================================================================================

    Friend Declare Function Bullet_CreateSimpleObject Lib "BulletWrapper.dll" ( _
                                                                                ByVal OffsetPosition As Vector3, _
                                                                                ByVal ShapeSize As Vector3, _
                                                                                ByVal ShapeType As ObjTypes, _
                                                                                Optional ByVal Mass As Single = 1, _
                                                                                Optional ByVal Margin As Single = 0.05, _
                                                                                Optional ByVal Activated As Boolean = True _
                                                                               ) As Int32

    ' ===============================================================================================
    '   STATIC OBJECT  /  MOVING OBJECT
    ' ===============================================================================================

    Private Declare Function Bullet_CreateObject Lib "BulletWrapper.dll" ( _
                                                                            ByVal pVertices As IntPtr, _
                                                                            ByVal pIndices As IntPtr, _
                                                                            ByVal numVertices As Int32, _
                                                                            ByVal NumberBytesPerVertex As Int32, _
                                                                            ByVal numFacs As Int32, _
                                                                            ByVal Is16BitIndices As Boolean, _
                                                                            ByVal WorldMatrix As Matrix, _
                                                                            ByVal CollisionShapeType As CollisionShapeTypes, _
                                                                            Optional ByVal Mass As Single = 1, _
                                                                            Optional ByVal Margin As Single = 0.05, _
                                                                            Optional ByVal Activated As Boolean = True _
                                                                          ) As Int32



    ' ===============================================================================================
    '   METHODS
    ' ===============================================================================================



    Friend Declare Sub Bullet_Object_ApplyCentralImpulse Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                                  ByVal Impulse As Vector3)

    Friend Declare Sub Bullet_Object_ApplyDamping Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                           ByVal Damping As Single)

    Friend Declare Sub Bullet_Object_ApplyTorqueImpulse Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                                 ByVal Torque As Vector3)

    Friend Declare Sub Bullet_Object_Translate Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                        ByVal Translation As Vector3)

    ' ===============================================================================================
    '   PROPERTIES SET
    ' ===============================================================================================

    Friend Declare Sub Bullet_Object_SetAdditionalDamping Lib "BulletWrapper.dll" ( _
                                                            ByVal ObjIndex As Int32, _
                                                            Optional ByVal AngularDampingThresholdSqr As Single = 0.8F, _
                                                            Optional ByVal AngularDampingFactor As Single = 0.8F, _
                                                            Optional ByVal LinearDampingThresholdSqr As Single = 0.02F, _
                                                            Optional ByVal LinearDampingFactor As Single = 0.8F)

    Friend Declare Sub Bullet_Object_SetActivationState Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                                 ByVal state As ActivationState)

    Friend Declare Sub Bullet_Object_SetDamping Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                         ByVal LinearDamping As Single, _
                                                                         ByVal AngularDamping As Single)

    Friend Declare Sub Bullet_Object_SetFriction Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                          ByVal Friction As Single)

    Friend Declare Sub Bullet_Object_SetKinematic Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                       ByVal kinematic As Boolean)

    Friend Declare Sub Bullet_Object_SetMatrix Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                        ByVal m As Matrix)

    Friend Declare Sub Bullet_Object_SetRestitution Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                             ByVal Restitution As Single)

    Friend Declare Sub Bullet_Object_SetSleepingThresholds Lib "BulletWrapper.dll" ( _
                                                        ByVal ObjIndex As Int32, _
                                                        Optional ByVal LinearThreshold As Single = 1.6F, _
                                                        Optional ByVal AngularThreshold As Single = 2.5F)

    ' ===============================================================================================
    '   PROPERTIES GET
    ' ===============================================================================================

    Friend Declare Function Bullet_Object_GetActivationState Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As ActivationState

    Friend Declare Function Bullet_Object_GetAngularDamping Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Single

    Friend Declare Function Bullet_Object_GetFriction Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Single

    Friend Declare Function Bullet_Object_GetLinearDamping Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Single

    Friend Declare Function Bullet_Object_GetLinearVelocity Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Vector3

    Friend Declare Function Bullet_Object_GetMass Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Single

    Friend Declare Function Bullet_Object_GetMatrix Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Matrix

    Friend Declare Function Bullet_Object_GetMatrixScaled Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Matrix

    Friend Declare Function Bullet_Object_GetRestitution Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Single



    Friend Declare Function Bullet_Object_GetPosition Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Vector3


    ' ================================================================================================
    '   SHOOT OBJECT
    ' ================================================================================================

    Friend Declare Sub Bullet_Object_Shoot Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                    ByVal Origin As Vector3, _
                                                                    ByVal Destination As Vector3, _
                                                                    ByVal ShootSpeed As Single)



    ' ================================================================================================
    '   CONSTRAINTS - POINT
    ' ================================================================================================

    Friend Declare Sub Bullet_Constraint_ObjectPoint Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                              ByVal PivotPoint As Vector3)

    Friend Declare Sub Bullet_Constraint_ObjectPointObject Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                    ByVal ObjIndex_B As Int32, _
                                                                                    ByVal PivotPoint_A As Vector3, _
                                                                                    ByVal PivotPoint_B As Vector3)

    ' ================================================================================================
    '   CONSTRAINTS - HINGE
    ' ================================================================================================

    Friend Declare Sub Bullet_Constraint_ObjectHinge_Matrix Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                     ByVal Matrix_A As Matrix, _
                                                                                     Optional ByVal useReferenceFrameA As Boolean = False)

    Friend Declare Sub Bullet_Constraint_ObjectHinge_PivotAndAxis Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                           ByVal Pivot_A As Vector3, _
                                                                                           ByVal Axis_A As Vector3, _
                                                                                           Optional ByVal useReferenceFrameA As Boolean = False)

    Friend Declare Sub Bullet_Constraint_ObjectHingeObject_Matrix Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                           ByVal ObjIndex_B As Int32, _
                                                                                           ByVal Matrix_A As Matrix, _
                                                                                           ByVal Matrix_B As Matrix, _
                                                                                           Optional ByVal useReferenceFrameA As Boolean = False)

    Friend Declare Sub Bullet_Constraint_ObjectHingeObject_PivotAndAxis Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                                 ByVal ObjIndex_B As Int32, _
                                                                                                 ByVal PivotPoint_A As Vector3, _
                                                                                                 ByVal PivotPoint_B As Vector3, _
                                                                                                 ByVal Axis_A As Vector3, _
                                                                                                 ByVal Axis_B As Vector3, _
                                                                                                 Optional ByVal useReferenceFrameA As Boolean = False)

    ' ================================================================================================
    '   CONSTRAINTS - SLIDER
    ' ================================================================================================

    Friend Declare Sub Bullet_Constraint_ObjectSlider Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                               ByVal Matrix_A As Matrix, _
                                                                               Optional ByVal useReferenceFrameA As Boolean = False)

    Friend Declare Sub Bullet_Constraint_ObjectSliderObject Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                     ByVal ObjIndex_B As Int32, _
                                                                                     ByVal Matrix_A As Matrix, _
                                                                                     ByVal Matrix_B As Matrix, _
                                                                                     Optional ByVal useReferenceFrameA As Boolean = False)


    ' ================================================================================================
    '   CONSTRAINTS - CONE-TWIST
    ' ================================================================================================

    Friend Declare Sub Bullet_Constraint_ObjectConeTwist Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                  ByVal Matrix_A As Matrix)

    Friend Declare Sub Bullet_Constraint_ObjectConeTwistObject Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
                                                                                        ByVal ObjIndex_B As Int32, _
                                                                                        ByVal Matrix_A As Matrix, _
                                                                                        ByVal Matrix_B As Matrix)

    ' ================================================================================================
    '   CONSTRAINTS - Generic6Dof - ToDo(complete with limits)
    ' ================================================================================================

    'Friend Declare Sub Bullet_Constraint_Object_Generic6Dof Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
    '                                                                                 ByVal Matrix_A As Matrix)

    'Friend Declare Sub Bullet_Constraint_Object_Generic6Dof_Object Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32, _
    '                                                                                        ByVal ObjIndex_B As Int32, _
    '                                                                                        ByVal Matrix_A As Matrix, _
    '                                                                                        ByVal Matrix_B As Matrix)



    ' ================================================================================================
    '   COLLISIONS
    ' ================================================================================================

    'Friend Declare Function Bullet_GetCollisions Lib "BulletWrapper.dll" (ByVal pCollisionArray As Int32, _
    '                                                                      Optional ByVal minCollision As Single = 0.001) As Int32

    'Friend Declare Function Bullet_GetCollidingObject Lib "BulletWrapper.dll" (ByVal ObjIndex_A As Int32) As Int32



    ' ================================================================================================
    '   VEHICLE
    ' ================================================================================================
    Private Declare Function Bullet_Vehicle_Create Lib "BulletWrapper.dll" ( _
                                                                                ByVal pVertices As IntPtr, _
                                                                                ByVal pIndices As IntPtr, _
                                                                                ByVal numVertices As Int32, _
                                                                                ByVal NumberBytesPerVertex As Int32, _
                                                                                ByVal numFacs As Int32, _
                                                                                ByVal Is16BitIndices As Boolean, _
                                                                                ByVal ShapeDelta As Vector3, _
                                                                                ByVal MassBarycenter As Vector3, _
                                                                                ByVal WorldMatrix As Matrix, _
                                                                                Optional ByVal Mass As Single = 100, _
                                                                                Optional ByVal Margin As Single = 0.05 _
                                                                            ) As Int32

    Friend Declare Sub Bullet_Vehicle_AddWheel Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                        ByVal Dx As Single, _
                                                                        ByVal Dy As Single, _
                                                                        ByVal Dz As Single, _
                                                                        ByVal Radius As Single, _
                                                                        ByVal SuspensionRestLength As Single, _
                                                                        ByVal IsFrontWheel As Boolean, _
                                                                        Optional ByVal DirectionX As Single = 0, _
                                                                        Optional ByVal DirectionY As Single = -0.5, _
                                                                        Optional ByVal DirectionZ As Single = 0, _
                                                                        Optional ByVal AxleX As Single = -0.5, _
                                                                        Optional ByVal AxleY As Single = 0, _
                                                                        Optional ByVal AxleZ As Single = 0)


    Friend Declare Sub Bullet_Vehicle_SetWheelTuning Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                              ByVal WheelIndex As Int32, _
                                                                              Optional ByVal SuspensionStiffness As Single = 20, _
                                                                              Optional ByVal SuspensionDamping As Single = 1.5, _
                                                                              Optional ByVal SuspensionCompression As Single = 1.0, _
                                                                              Optional ByVal WheelFriction As Single = 6.0, _
                                                                              Optional ByVal RollInfluence As Single = 0.1, _
                                                                              Optional ByVal MaxSuspensionTravelCm As Single = 999)

    Friend Declare Function Bullet_Vehicle_GetMatrix Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Matrix

    Friend Declare Function Bullet_Vehicle_GetMatrixScaled Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Matrix

    Private Declare Sub Bullet_Vehicle_GetWheelMatrices Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                                 ByVal pMatrices As IntPtr)

    Friend Declare Sub Bullet_Vehicle_SetEngineForce Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                              ByVal WheelIndex As Int32, _
                                                                              ByVal Force As Single)

    Friend Declare Sub Bullet_Vehicle_SetBrakingForce Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                               ByVal WheelIndex As Int32, _
                                                                               ByVal Force As Single)

    Friend Declare Sub Bullet_Vehicle_SetSteeringValue Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                                ByVal WheelIndex As Int32, _
                                                                                ByVal SteeringValue As Single)







    ' ===============================================================================================
    '   SOFT BODY - TUNE-PARAMS
    ' ===============================================================================================


    Friend Declare Sub Bullet_SoftObj_TuneParams Lib "BulletWrapper.dll" ( _
                                                                        ByVal ObjIndex As Int32, _
                                                                        Optional ByVal LinearStiffness As Single = 0.3, _
                                                                        Optional ByVal AngularStiffness As Single = 0.3, _
                                                                        Optional ByVal Damping As Single = 0.01, _
                                                                        Optional ByVal DynamicFriction As Single = 0.02, _
                                                                        Optional ByVal RigidContactHardness As Single = 0.5, _
                                                                        Optional ByVal KineticContactHardness As Single = 0.3, _
                                                                        Optional ByVal SoftContactHardness As Single = 0.3, _
                                                                        Optional ByVal AnchorHardness As Single = 0.1 _
                                                                      )

    Friend Declare Sub Bullet_SoftObj_SetDamping Lib "BulletWrapper.dll" ( _
                                                                          ByVal ObjIndex As Int32, _
                                                                          Optional ByVal Damping As Single = 0.01 _
                                                                          )

    Friend Declare Function Bullet_SoftObj_GetDamping Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Single

    Friend Declare Sub Bullet_SoftObj_TuneAeroParams Lib "BulletWrapper.dll" ( _
                                                                    ByVal ObjIndex As Int32, _
                                                                    ByVal WindVelocity As Vector3, _
                                                                    Optional ByVal AeroLift As Single = 0.1, _
                                                                    Optional ByVal AeroDrag As Single = 0.01 _
                                                                  )

    Friend Declare Sub Bullet_SoftObj_TuneVolumeParams Lib "BulletWrapper.dll" ( _
                                                                                ByVal ObjIndex As Int32, _
                                                                                Optional ByVal VolumeStiffness As Single = 0.3, _
                                                                                Optional ByVal PressureCoeff As Single = 0.1, _
                                                                                Optional ByVal VolumeConversationCoeff As Single = 0.5, _
                                                                                Optional ByVal PoseMatchingCoeff As Single = 0.5, _
                                                                                Optional ByVal MaximumVolumeRatioForPose As Single = 1.0 _
                                                                               )


    Friend Declare Sub Bullet_SoftObj_TuneSolverParams Lib "BulletWrapper.dll" ( _
                                                                                ByVal ObjIndex As Int32, _
                                                                                Optional ByVal VelocitiesCorrectionFactor As Single = 1.0, _
                                                                                Optional ByVal VelocitiesSolverIterations As Int32 = 0, _
                                                                                Optional ByVal PositionsSolverIterations As Int32 = 2, _
                                                                                Optional ByVal DriftSolverIterations As Int32 = 0, _
                                                                                Optional ByVal ClusterSolverIterations As Int32 = 4, _
                                                                                Optional ByVal Timescale As Single = 1.0 _
                                                                               )

    Friend Declare Sub Bullet_SoftObj_UseClusters Lib "BulletWrapper.dll" ( _
                                                                            ByVal ObjIndex As Int32, _
                                                                            Optional ByVal ClusterCount As Int32 = 70, _
                                                                            Optional ByVal SoftVsRigidHardness As Single = 0.3, _
                                                                            Optional ByVal SoftVsRigidImpulseSplit As Single = 0.1, _
                                                                            Optional ByVal SoftVsSoftHardness As Single = 0.1, _
                                                                            Optional ByVal SoftVsSoftImpulseSplit As Single = 0.1, _
                                                                            Optional ByVal SoftVsKineticHardness As Single = 0.1, _
                                                                            Optional ByVal SoftVsKineticImpulseSplit As Single = 0.1 _
                                                                          )




    ' ===============================================================================================
    '   SOFT BODY - btSoftBody
    ' ===============================================================================================


    Private Declare Function Bullet_SoftObj_CreateSoftObject Lib "BulletWrapper.dll" ( _
                                                                                    ByVal pVertices As IntPtr, _
                                                                                    ByVal pIndices As IntPtr, _
                                                                                    ByVal numVertices As Int32, _
                                                                                    ByVal NumberBytesPerVertex As Int32, _
                                                                                    ByVal numFacs As Int32, _
                                                                                    ByVal Is16BitIndices As Boolean, _
                                                                                    ByVal WorldMatrix As Matrix, _
                                                                                    Optional ByVal Mass As Single = 1, _
                                                                                    Optional ByVal Margin As Single = 0.05, _
                                                                                    Optional ByVal BendingConstraintsDist As Int32 = 2 _
                                                                                  ) As Int32


    Friend Declare Function Bullet_SoftObj_CreateCloth Lib "BulletWrapper.dll" ( _
                                                                                ByVal Size_x As Single, _
                                                                                ByVal Size_z As Single, _
                                                                                Optional ByVal NodeCount_x As Int32 = 11, _
                                                                                Optional ByVal NodeCount_z As Int32 = 11, _
                                                                                Optional ByVal Pos_x As Single = 0, _
                                                                                Optional ByVal Pos_y As Single = 0, _
                                                                                Optional ByVal Pos_z As Single = 0, _
                                                                                Optional ByVal Yaw As Single = 0, _
                                                                                Optional ByVal Pitch As Single = 0, _
                                                                                Optional ByVal Roll As Single = 0, _
                                                                                Optional ByVal FixedSides As Sides = Sides.none, _
                                                                                Optional ByVal TotalMass As Single = 100, _
                                                                                Optional ByVal Margin As Single = 0.2, _
                                                                                Optional ByVal BendingConstraintsDist As Int32 = 2) As Int32


    Friend Declare Function Bullet_SoftObj_CreateRope Lib "BulletWrapper.dll" ( _
                                                                            ByVal FromPoint As Vector3, _
                                                                            ByVal ToPoint As Vector3, _
                                                                            Optional ByVal SegmentCount As Int32 = 31, _
                                                                            Optional ByVal FixedSides As RopeSides = RopeSides.none, _
                                                                            Optional ByVal TotalMass As Single = 1, _
                                                                            Optional ByVal Margin As Single = 0.2, _
                                                                            Optional ByVal BendingConstraintsDist As Integer = 2 _
                                                                           ) _
                                                                           As Int32



    Friend Declare Function Bullet_SoftObj_GetNodesCount Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Int32


    Friend Declare Sub Bullet_SoftObj_GetNodes Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                        ByVal pNodes As IntPtr, _
                                                                        Optional ByVal NodesCount As Int32 = -1)

    Friend Declare Function Bullet_SoftObj_GetFacesCount Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32) As Int32


    Friend Declare Sub Bullet_SoftObj_GetFaces Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                        ByVal pFaces As IntPtr)



    Friend Declare Sub Bullet_SoftObj_ApplyForce Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                          ByVal Force As Vector3, _
                                                                          ByVal Node As Int32)

    Friend Declare Sub Bullet_SoftObj_Cut Lib "BulletWrapper.dll" (ByVal ObjIndex As Int32, _
                                                                   ByVal Position As Vector3, _
                                                                   ByVal Radius As Single, _
                                                                   Optional ByVal Accuracy As Single = 0.0001)



    ' -----------------------------------------------------------------------------------------------------
    '  A negative node number specifyes a node near to the "end" side  
    '   ( "0" is the first node and "-1" is the last node )
    ' -----------------------------------------------------------------------------------------------------

    Friend Declare Sub Bullet_SoftObj_LinkToRigid Lib "BulletWrapper.dll" (ByVal SoftBodyIndex As Int32, _
                                                                           ByVal Node As Int32, _
                                                                           ByVal RigidBodyIndex As Int32, _
                                                                           Optional ByVal DisableCollisionBetweenLinkedBodies As Boolean = False)


    Friend Sub Bullet_SoftObj_LinkToSoft(ByVal SoftBodyIndex1 As Int32, _
                                         ByVal Node1 As Int32, _
                                         ByVal SoftBodyIndex2 As Int32, _
                                         ByVal Node2 As Int32, _
                                         Optional ByVal DisableCollisionBetweenLinkedBodies As Boolean = False)

        Dim p1 As Vector3 = Bullet_SoftObj_GetNodePosition(SoftBodyIndex1, Node1)
        Dim p2 As Vector3 = Bullet_SoftObj_GetNodePosition(SoftBodyIndex2, Node2)
        Dim p3 As Vector3 = (p1 + p2) * 0.5

        App_CreateSimpleObject(p2, VEC3(0.001, 0.001, 0.001), DrawTypes.INVISIBLE, ObjTypes.BOX, SoundTypes.NoSound, 1)
        Dim r1 As Int32 = LastBtIndex

        'Dim r1 As Int32 = Bullet_CreateSimpleObject(p2, VEC3(0.1, 0.1, 0.1), DrawTypes.INVISIBLE, ObjTypes.BOX, 1)

        Bullet_SoftObj_LinkToRigid(SoftBodyIndex1, Node1, r1, True)
        Bullet_SoftObj_LinkToRigid(SoftBodyIndex2, Node2, r1, True)
    End Sub


    ' ===============================================================================================
    '   
    ' ===============================================================================================

    Friend Enum CollisionShapeTypes
        BOX = 0
        'SPHERE = 8
        'CAPSULE = 10
        'CONE = 11
        'CYLINDER = 13
        CONVEX_HULL = 50
        'COMPOUND = 51
        'MOVING_MESH = 52
        TRIANGLE_MESH_SHAPE = 99
        'HeightField = 100
        'ScaledStaticMesh = 101
    End Enum

    Friend Enum DrawTypes As Int32
        SOFT_OBJECT
        CAMERA
        GENERIC
        TERRAIN_CASTLE
        TERRAIN_FLAT
        TERRAIN_MESH
        MILL
        ROPE
        CLOTH
        CAR
        EIFFEL
        CASTLE
        BOX
        BALL
        BARREL
        CHROME
        INVISIBLE
    End Enum

    Friend Enum ObjTypes As Int32
        BOX = 0
        SPHERE = 8
        CAPSULE = 10
        CONE = 11
        CYLINDER = 13

        CAR = -1
        ROPE = -2
        CLOTH = -3
        SOFT_OBJECT = -4

        GENERIC = -99
    End Enum

    Friend Enum ActivationState
        ACTIVE_TAG = 1
        ISLAND_SLEEPING = 2
        WANTS_DEACTIVATION = 3
        DISABLE_DEACTIVATION = 4
        DISABLE_SIMULATION = 5
    End Enum

    Friend Enum Sides
        none = 0
        corner00 = 1
        corner01 = 2
        corner10 = 4
        corner11 = 8
        all_corners = 15
        upper_middle = 16
        left_middle = 32
        right_middle = 64
        lower_middle = 128
        center = 256
    End Enum

    Friend Enum RopeSides
        none = 0
        startSide = 1
        endSide = 2
    End Enum


    Friend Bullet_Paused As Boolean = False

    Private DeactivationSpeed As Single = 0

    ' ===============================================================================================
    '   OPEN - CLOSE
    ' ===============================================================================================

    Friend Sub Bullet_Open()
        Dim size As Single = 1000
        Bullet_Init(VEC3(-size, -size, -size), _
                           VEC3(size, size, size), _
                           VEC3(0, -9.8, 0))
    End Sub

    Friend Sub Bullet_Close()
        Bullet_Thread_Stop()
        ' ------------------------------- Bullet_DestroyAll produces an error on exit if debugging from C++
        Bullet_Destroy()
    End Sub



    ' ===============================================================================================
    '   BULLET THREAD
    ' ===============================================================================================

    Friend Sub Bullet_Thread_Start()
        If Bullet_Thread Is Nothing Then
            Bullet_Thread = New Thread(AddressOf ThreadProc)
            'Bullet_Thread.Priority = System.Threading.ThreadPriority.Highest
            Bullet_Thread.Start()
        End If
    End Sub

    Friend Sub Bullet_Thread_Stop()
        If Bullet_Thread IsNot Nothing Then
            Bullet_Thread_Exit = True
            Do
                Thread.Sleep(1)
            Loop Until Bullet_Thread_IsStopped
            Bullet_Thread = Nothing
        End If
    End Sub


    Private Bullet_Thread As Thread = Nothing
    Private Bullet_Thread_Exit As Boolean
    Private Bullet_Thread_IsStopped As Boolean

    Friend Sub ThreadProc()

        Bullet_Thread_IsStopped = False
        Bullet_Thread_Exit = False

        Dim elapsed As Single
        Dim threadTimer As PrecisionTimer = New PrecisionTimer
        Do
            ' ------------------------------------------------------------------------- 1
            Do
                elapsed = CSng(threadTimer.GetPartialTime())
                If elapsed > 0.035 Then Exit Do
                Thread.Sleep(1)
            Loop
            threadTimer.StartTimer()
            'Debug.Print(elapsed.ToString)
            Bullet_Step(elapsed, 5)

            Thread.Sleep(0)

            ' ------------------------------------------------------------------------- 2
            'Thread.Sleep(30)
            'elapsed = CSng(threadTimer.GetTime())
            'threadTimer.StartTimer()
            ''Debug.Print(elapsed.ToString)
            'Bullet_StepSimulation(elapsed, 3)


        Loop Until Bullet_Thread_Exit

        Bullet_Thread_IsStopped = True
    End Sub




    ' ===============================================================================================
    '   STEP SIMULATION
    ' ===============================================================================================

    Friend Sub Bullet_Step(ByVal Interval As Single)

        If Bullet_Paused Then

            Bullet_Thread_Stop()
            ' ---------------------------------------------------------------- stop also the data-update
            Return

        Else
            If DEBUG_UseThreads Then
                Bullet_Thread_Start()
            Else
                Bullet_Thread_Stop()

                ' ---------------------------------------------------------------- do the BulletStep
                'Bullet_StepSimulation(Interval, 1)
                Bullet_Step(Interval, 2)
                'Bullet_StepSimulation(Interval, 2, Interval / 2)
                'Bullet_StepSimulation(Interval, 3, 0.015)
                'Bullet_StepSimulation(Interval, 9, 0.01)

                'If Interval < 0.04 Then
                '    Bullet_StepSimulation(Interval, 3, 0.01)
                'Else
                '    Bullet_StepSimulation(Interval, 2)
                '    'Beep()
                'End If

            End If
        End If





        Dim CollisionCount As Int32
        Dim NumBulletObjects As Int32
        NumBulletObjects = Bullet_GetNumObjects()

        'System.Threading.Thread.CurrentThread.Priority = System.Threading.ThreadPriority.Lowest
        'System.Diagnostics.Process.GetCurrentProcess().PriorityClass = ProcessPriorityClass.BelowNormal


        ' ---------------------
        '  test falling times
        ' ---------------------
        '   100 m    4.51 sec
        '    50 m    3.19 sec
        '    30 m    2.47 sec
        '    20 m    2.02 sec
        '    10 m    1.43 sec





        ' -------------------------------------------------------------------- positions and rotations
        Dim newGlobalSpeed As Vector3
        Dim newLocalSpeed As Vector3
        For i As Int32 = 0 To UBound(Objects_3D)
            With Objects_3D(i)

                If .ObjType = ObjTypes.CLOTH Or .ObjType = ObjTypes.SOFT_OBJECT Then
                    Bullet_SoftObj_GetNodes(.BulletIndex, VarPtr(.SoftObjNodes), .SoftObjNodes.Length)
                    TranferSoftObjNodesToMesh(.SoftObjNodes, .MeshX.msh)
                End If

                If .ObjType = ObjTypes.ROPE Then
                    Bullet_SoftObj_GetNodes(.BulletIndex, VarPtr(.SoftObjNodes), .SoftObjNodes.Length)
                    TranferRopeNodesToMesh(.SoftObjNodes, .RopeFaces, .RopeRadius, .MeshX.msh)
                End If

                .Activation = Bullet_Object_GetActivationState(.BulletIndex)
                .IsActive = (.Activation And ActivationState.ACTIVE_TAG) <> 0 Or .Activation = ActivationState.DISABLE_DEACTIVATION
                If Not .IsActive Then Continue For



                'Dim t As PrecisionTimer = New PrecisionTimer
                If .IsMovingObject Then
                    If .ObjType = ObjTypes.CAR Then

                        Dim ShapeDelta As Vector3 = VEC3(0, 0.1, 0)

                        .MeshX.mtx = Matrix.Translation(ShapeDelta) * Bullet_Vehicle_GetMatrixScaled(.BulletIndex)

                        Bullet_Vehicle_GetWheelMatrices(.BulletIndex, VarPtr(.WheelMatrix))

                        PlayCarMotor(i)

                    Else
                        ' -------------------------------------------------------- positions, rotations and scale
                        .MeshX.mtx = Bullet_Object_GetMatrixScaled(.BulletIndex)

                        'If i = CameraIndex Then
                        '    'Beep()
                        '    'Debug.Print(Matrix_GetVectorPos(.MeshX.mtx).ToString)
                        'End If

                    End If


                    ' ------------------------------------------------------------ global and local speed 
                    newGlobalSpeed = Bullet_Object_GetLinearVelocity(.BulletIndex)
                    newLocalSpeed = Vector3.TransformNormal(newGlobalSpeed, Matrix.Invert(.MeshX.mtx))

                    ' ------------------------------------------------------------ global and local acc
                    .GlobalAcc = (newGlobalSpeed - .GlobalSpeed) * k_FPS
                    .LocalAcc = (newLocalSpeed - .LocalSpeed) * k_FPS
                    .GlobalSpeed = newGlobalSpeed
                    .LocalSpeed = newLocalSpeed
                    .Position = Matrix_GetVectorPos(.MeshX.mtx)

                    'If i = CarActive Then
                    '    Form1.Label99.Text = .LocalSpeed.Z.ToString("0.000")
                    'End If

                    ' -------------------------------------------------------- collisions and sounds
                    If .GlobalAcc.Length > 25 Then

                        ' ---------------------------------------------------- slow down spheres
                        If .ObjType = ObjTypes.SPHERE Then

                            Dim c As Int32 = DX_FindCollidingObject(i)
                            'Debug.Print(c.ToString)

                            If c >= 0 Then
                                If Objects_3D(c).ObjType = ObjTypes.CLOTH Then
                                    If Not .Shooted Then
                                        Dim mass As Single = Bullet_Object_GetMass(i)
                                        Bullet_Object_ApplyCentralImpulse(.BulletIndex, .GlobalSpeed * -mass * 0.25)
                                        CollisionCount += 1
                                        PlayObjectSound(c)
                                        Continue For
                                    End If
                                End If
                            End If

                        End If

                        CollisionCount += 1
                        PlayObjectSound(i)
                    End If

                    ' -------------------------------------------------------- sleep
                    If .GlobalSpeed.Length < DeactivationSpeed Then
                        Bullet_Object_SetActivationState(.BulletIndex, ActivationState.WANTS_DEACTIVATION)
                    End If

                End If


            End With
        Next

        If DEBUG_ViewMainParams Then
            ' -------------------------------------------------------------------- show number of moving objects
            Form1.lbl_objects.Text = "   Objects " & NumBulletObjects.ToString
            Form1.lbl_collisions.Text = "Collisions " & CollisionCount.ToString
            SetVisibleLabels(True)
        Else
            SetVisibleLabels(False)
        End If

    End Sub

    Private Sub SetVisibleLabels(ByVal visible As Boolean)
        If Form1.Label99.Visible <> visible Then
            Form1.Label99.Visible = visible
            Form1.lbl_objects.Visible = visible
            Form1.lbl_collisions.Visible = visible
            Form1.lbl_TimePhysics.Visible = visible
            Form1.lbl_TimeTotal.Visible = visible
            Form1.lbl_Fps.Visible = visible
            Form1.lbl_FpsStab.Visible = visible
        End If
    End Sub

    Friend Sub Bullet_SetDeactivationSpeed(Optional ByVal speed As Single = 0.05)
        DeactivationSpeed = speed
    End Sub


    Private Function DX_FindCollidingObject(ByVal ObjIndex As Int32) As Int32
        With Objects_3D(ObjIndex)
            RayPick(.Position, .GlobalSpeed, ObjIndex)
        End With
        Return Pick_GetPickedObjIndex()
    End Function



    ' ===============================================================================================
    '   CREATE OBJECTS FROM MESH
    ' ===============================================================================================

    Friend Function Bullet_CreateSimpleObjectFromMesh(ByVal objMesh As Mesh, ByVal mtx As Matrix) As Int32
        Dim bbmin As Vector3
        Dim bbmax As Vector3
        DX_ComputeBoundingBoxFromMesh(objMesh, bbmin, bbmax)

        LastBtIndex = Bullet_CreateSimpleObject(VEC3(mtx.M41, mtx.M42, mtx.M43), _
                                                         bbmax - bbmin, _
                                                         ObjTypes.BOX, _
                                                         20, _
                                                         , _
                                                         True)
        Return LastBtIndex
    End Function


    Friend Function Bullet_CreateObjectFromMesh(ByVal objMesh As Mesh, _
                                                ByVal WorldMatrix As Matrix, _
                                                Optional ByVal CollisionShapeType As CollisionShapeTypes = CollisionShapeTypes.TRIANGLE_MESH_SHAPE, _
                                                Optional ByVal Mass As Single = 1, _
                                                Optional ByVal Margin As Single = 0.05, _
                                                Optional ByVal Activated As Boolean = True) As Int32

        Dim gs1 As GraphicsStream = objMesh.VertexBuffer.Lock(0, 0, LockFlags.ReadOnly)
        Dim gs2 As GraphicsStream = objMesh.IndexBuffer.Lock(0, 0, LockFlags.ReadOnly)
        LastBtIndex = Bullet_CreateObject(gs1.InternalData, gs2.InternalData, _
                                             objMesh.NumberVertices, _
                                             objMesh.NumberBytesPerVertex, _
                                             objMesh.NumberFaces, _
                                             objMesh.IndexBuffer.Description.Is16BitIndices, _
                                             WorldMatrix, _
                                             CollisionShapeType, _
                                             Mass, _
                                             Margin, _
                                             Activated)
        objMesh.VertexBuffer.Unlock()
        objMesh.IndexBuffer.Unlock()
        Return LastBtIndex
    End Function










    ' ===============================================================================================
    '   VEHICLE
    ' ===============================================================================================

    Friend Function Bullet_CreateVehicle(ByVal objMesh As Mesh, _
                                         ByVal ShapeDelta As Vector3, _
                                         ByVal MassBarycenter As Vector3, _
                                         ByVal WorldMatrix As Matrix, _
                                         Optional ByVal Mass As Single = 100, _
                                         Optional ByVal Margin As Single = 0.05) As Int32

        Dim gs1 As GraphicsStream
        Dim gs2 As GraphicsStream
        gs1 = objMesh.VertexBuffer.Lock(0, 0, LockFlags.ReadOnly)
        gs2 = objMesh.IndexBuffer.Lock(0, 0, LockFlags.ReadOnly)
        LastBtIndex = Bullet_Vehicle_Create(gs1.InternalData, _
                                                   gs2.InternalData, _
                                                   objMesh.NumberVertices, _
                                                   objMesh.NumberBytesPerVertex, _
                                                   objMesh.NumberFaces, _
                                                   objMesh.IndexBuffer.Description.Is16BitIndices, _
                                                   ShapeDelta, _
                                                   MassBarycenter, _
                                                   WorldMatrix, _
                                                   Mass, _
                                                   Margin)

        objMesh.VertexBuffer.Unlock()
        objMesh.IndexBuffer.Unlock()
        Return LastBtIndex
    End Function




    ' ===============================================================================================
    '   SOFT BODIES
    ' ===============================================================================================

    Private Sub TranferSoftObjNodesToMesh(ByRef SoftNodes() As Vector3, ByRef m As Mesh)
        Dim FVFsize As Int32 = VertexInformation.GetFormatSize(m.VertexFormat)
        Dim gs As GraphicsStream
        For j As Int32 = 0 To m.NumberVertices - 1
            gs = m.VertexBuffer.Lock(j * FVFsize, 12, LockFlags.None)
            gs.Write(SoftNodes(j))
            m.VertexBuffer.Unlock()
        Next
        If (m.VertexFormat And VertexFormats.Normal) <> 0 Then
            m.ComputeNormals()
        End If
    End Sub


    Private Sub TranferRopeNodesToMesh(ByRef SoftNodes() As Vector3, ByVal nFaces As Int32, ByVal Radius As Single, ByRef m As Mesh)
        Dim FVFsize As Int32 = VertexInformation.GetFormatSize(m.VertexFormat)
        Dim gs As GraphicsStream
        Dim StepAngle As Single = 360 * KRAD / nFaces
        Dim v0, v1, v2, v3 As Vector3
        v0 = SoftNodes(0)
        For i As Int32 = 1 To SoftNodes.Length
            ' --------------------------------------------------------- last segment direction used two times
            If i < SoftNodes.Length Then
                v1 = SoftNodes(i)
                v2 = v1 - v0
            End If
            ' --------------------------------------------------------- calculate  v3 = ortho(v2) * radius
            v3 = New Vector3(v2.Z - v2.Y, v2.X - v2.Z, v2.Y - v2.X)
            v3 = Radius * Vector3.Normalize(v3)
            ' --------------------------------------------------------- set vertex positions
            For j As Int32 = 0 To nFaces
                gs = m.VertexBuffer.Lock(((i - 1) * (nFaces + 1) + j) * FVFsize, 12, LockFlags.None)
                gs.Write(v0 + v3)
                m.VertexBuffer.Unlock()
                v3 = Vector3.TransformNormal(v3, Matrix.RotationAxis(v2, StepAngle))
            Next
            v0 = v1
        Next
        m.ComputeNormals()
    End Sub


    ' ======================================================================================================
    ' - UNUSED -
    '   GET-MESH can rebuild the mesh after a CUT-MESH
    '   but there are too many problems with: "Texture-Coords" and with: "Draw separate triangles" 
    ' ------------------------------------------------------------------------------------------------------
    '   This function ( or a similar rigid body function ) can be used to see bullet objects
    ' ======================================================================================================

    Friend Function Bullet_SoftObj_GetMesh(ByVal obj As Object_3D) As Mesh
        ' 
        Dim numVerts As Integer = Bullet_SoftObj_GetNodesCount(obj.BulletIndex)
        Dim numFaces As Integer = Bullet_SoftObj_GetFacesCount(obj.BulletIndex)
        ' ----------------------------------------------------------------------- mesh
        Dim m As Mesh = New Mesh(numFaces, _
                                 numVerts, _
                                 MeshFlags.Managed, _
                                 CustomVertex.PositionNormalTextured.Format, _
                                 DX_Device)

        ' ----------------------------------------------------------------------- vertices
        'Dim gs As GraphicsStream = m.VertexBuffer.Lock(0, 0, LockFlags.None)
        'For h As Int32 = 0 To numVerts - 1
        '    gs.Write(New CustomVertex.PositionNormalTextured(0, 0, 0, 0, 1, 0, 0, 0))
        'Next
        'm.VertexBuffer.Unlock()

        ReDim obj.SoftObjNodes(numVerts - 1)
        Bullet_SoftObj_GetNodes(obj.BulletIndex, VarPtr(obj.SoftObjNodes), numVerts)
        TranferSoftObjNodesToMesh(obj.SoftObjNodes, m)

        Dim gs As GraphicsStream = m.VertexBuffer.Lock(0, 0, LockFlags.None)
        Dim posMin As Vector3 = obj.SoftObjNodes(0)
        Dim posMax As Vector3 = obj.SoftObjNodes(numVerts - 1)
        Dim sizeX As Single = posMax.X - posMin.X
        Dim sizeZ As Single = posMax.Z - posMin.Z
        For h As Int32 = 0 To numVerts - 1
            Dim pos As Vector3 = obj.SoftObjNodes(h)
            Dim u As Single = (pos.X - posMin.X) / sizeX
            Dim v As Single = (pos.Z - posMin.Z) / sizeZ
            gs.Write(New CustomVertex.PositionNormalTextured(0, 0, 0, 0, 1, 0, u, 1 - v))
        Next
        m.VertexBuffer.Unlock()


        ' ----------------------------------------------------------------------- indices
        Dim index32(3 * numFaces - 1) As Int32
        Dim index16(3 * numFaces - 1) As Int16
        Bullet_SoftObj_GetFaces(obj.BulletIndex, VarPtr(index32))
        For i As Int32 = 0 To numFaces * 3 - 1
            index16(i) = CShort(index32(i))
        Next
        m.IndexBuffer.SetData(index16, 0, LockFlags.None)
        '
        Return m
    End Function





    Friend Function Bullet_CreateSoftObjectFromMesh(ByVal objMesh As Mesh, _
                                                    ByVal WorldMatrix As Matrix, _
                                                    Optional ByVal Mass As Single = 1, _
                                                    Optional ByVal Margin As Single = 0.05, _
                                                    Optional ByVal BendingConstraintsDist As Int32 = 2) As Int32

        Dim gs1 As GraphicsStream = objMesh.VertexBuffer.Lock(0, 0, LockFlags.ReadOnly)
        Dim gs2 As GraphicsStream = objMesh.IndexBuffer.Lock(0, 0, LockFlags.ReadOnly)
        LastBtIndex = Bullet_SoftObj_CreateSoftObject(gs1.InternalData, _
                                                                gs2.InternalData, _
                                                                objMesh.NumberVertices, _
                                                                objMesh.NumberBytesPerVertex, _
                                                                objMesh.NumberFaces, _
                                                                objMesh.IndexBuffer.Description.Is16BitIndices, _
                                                                WorldMatrix, _
                                                                Mass, _
                                                                Margin, _
                                                                BendingConstraintsDist)

        objMesh.VertexBuffer.Unlock()
        objMesh.IndexBuffer.Unlock()
        Return LastBtIndex
    End Function




    Friend Function Bullet_SoftObj_GetNodePosition(ByVal BtIndex As Int32, ByVal Node As Int32) As Vector3
        Dim numVerts As Integer = Bullet_SoftObj_GetNodesCount(BtIndex)
        Dim nodes(numVerts) As Vector3
        Bullet_SoftObj_GetNodes(BtIndex, VarPtr(nodes), numVerts)
        If Node >= 0 And Node < numVerts Then
            Return nodes(Node)
        Else
            Return New Vector3(0, 0, 0)
        End If
    End Function

    Friend Function Bullet_SoftObj_FindNearestNode(ByVal BtIndex As Int32, ByVal Position As Vector3) As Int32
        Dim numVerts As Integer = Bullet_SoftObj_GetNodesCount(BtIndex)
        Dim nodes(numVerts) As Vector3
        Bullet_SoftObj_GetNodes(BtIndex, VarPtr(nodes), numVerts)
        Dim minDist As Single = 3.0E+33F
        Dim node As Int32
        For i As Int32 = 0 To numVerts - 1
            Dim d As Single = (Position - nodes(i)).Length
            If d < minDist Then
                node = i
                minDist = d
            End If
        Next
        Return node
    End Function


End Module
