﻿
Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D
Imports D3D = Microsoft.DirectX.Direct3D

' DirectX DLLs are to be added from GAC 2902 with SpecificVersion=True and CopyLocal=True
' Project/References/Add/Browse/
' C:\WINDOWS\assembly\GAC\Microsoft.DirectX\1.0.2902.0__31bf3856ad364e35\Microsoft.DirectX.dll
' C:\WINDOWS\assembly\GAC\Microsoft.DirectX.Direct3D\1.0.2902.0__31bf3856ad364e35\Microsoft.DirectX.Direct3D.dl
' C:\WINDOWS\assembly\GAC\Microsoft.DirectX.Direct3DX\1.0.2902.0__31bf3856ad364e35\Microsoft.DirectX.Direct3DX.dll
' C:\WINDOWS\assembly\GAC\Microsoft.DirectX.DirectSound\1.0.2902.0__31bf3856ad364e35\Microsoft.DirectX.DirectSound.dll

Module DirectX

    Friend ControlType As ControlTypes

    Friend LookAtObjectIndex As Int32

    Friend DX_Device As Device
    Friend DX_Wnd As Control

    ' ------------------------------------------------- Params Default without INI
    Friend DEVICE_BackBufferWidth As Int32 = 0
    Friend DEVICE_BackBufferHeight As Int32 = 0
    Friend DEVICE_Multisamples As Int32 = 4
    Friend DEVICE_HardwareVertexProcessing As Boolean = True

    Friend DEBUG_ViewMainParams As Boolean = False
    Friend DEBUG_ViewShadowMap As Boolean = False
    Friend DEBUG_ViewFinalMap As Boolean = False
    Friend DEBUG_DisableShadows As Boolean = False
    Friend DEBUG_DisableSounds As Boolean = False
    Friend DEBUG_HiQuality As Boolean = True
    Friend DEBUG_DrawWireFrame As Boolean = False
    Friend DEBUG_RunInWindow As Boolean = True
    Friend DEBUG_UseThreads As Boolean = False
    Friend DEBUG_ShowMovements As Boolean = False
    Friend DEBUG_ViewDebugParams As Boolean = False

    Friend f_ViewAspect As Single = 1
    Friend f_ShootSpeed As Single

    ' --------------------------------- this PICK comparation can fail if farplane / nearplane > 100k
    Friend f_NearPlane As Single = 0.1   ' 0.1
    Friend f_FarPlane As Single = 9000    ' 9000

    Friend mtx_World As Matrix
    Friend mtx_CameraView As Matrix
    Friend mtx_CameraProj As Matrix


    Friend tex_AmbientPosZ As Texture
    Friend tex_AmbientPosX As Texture
    Friend tex_AmbientNegZ As Texture
    Friend tex_AmbientNegX As Texture
    Friend tex_AmbientNegY As Texture
    Friend tex_AmbientPosY As Texture


    Friend Enum ControlTypes
        MoveCamera
        MoveVehicleWithInternalCamera
        MoveVehicle
        AnimateVehicles
    End Enum


    Friend Structure CameraStruct
        Dim Lens As Single
        Dim Pos As Vector3
        Dim At As Vector3
        Dim Right As Vector3
        Dim Dir As Vector3
        Dim Up As Vector3
    End Structure
    Friend Camera As CameraStruct

    Friend Const KRAD As Single = Math.PI / 180

    Friend Present_Params As PresentParameters




    ' ===============================================================================================
    '   OPEN / CLOSE / STEP
    ' ===============================================================================================

    Friend Sub DX_Open(ByVal Wnd As Control)
        DX_Wnd = Wnd
        f_ViewAspect = CSng(Wnd.Width / Wnd.Height)
        CreateDevice(Wnd.Handle)
        DX_LoadMeshAndTextures()
        DX_SetCamera()
        DX_InitMaterials()
        DX_InitLights()
        Effects_Init(False)
    End Sub

    Friend Sub DX_Resize(ByVal Wnd As Control)
        If DX_Device Is Nothing Then Exit Sub
        ' -------------------------------------------------------------- 
        f_ViewAspect = CSng(DX_Wnd.Width / DX_Wnd.Height)
    End Sub

    Friend Sub DX_Close()
        StopTimers()
        For i As Int32 = 0 To UBound(Materials)
            Materials(i) = Nothing
        Next
        For i As Int32 = 0 To UBound(msh_Array)
            If msh_Array(i) IsNot Nothing Then
                msh_Array(i).Dispose()
                msh_Array(i) = Nothing
            End If
        Next
        'For i As Int32 = 0 To UBound(Textures)
        '    Textures(i).Dispose()
        '    Textures(i) = Nothing
        'Next
        Effects_Close()
        DX_Device = Nothing
    End Sub




    Friend Sub DX_ResetDevice()
        Do
            Try
                DX_Device.Reset(Present_Params)
            Catch
            End Try
        Loop Until DX_Device.CheckCooperativeLevel
        DX_InitLights()
        'App_StartScene()
        System.Threading.Thread.Sleep(500)
        Form1.Focus()
        Form1.Refresh()
    End Sub

    Friend Sub DX_ReinitDevice()
        DX_Close()
        DX_Open(Form1)
        App_StartScene()
        StartTimers()

        'Present_Params.BackBufferWidth = 1024
        'Present_Params.BackBufferWidth = 768
        'Present_Params.BackBufferCount = 1
        'DX_ResetDevice()
    End Sub



    ' ===============================================================================================
    '   DEVICE
    ' ===============================================================================================


    Private Sub CreateDevice(ByVal WindowHandle As IntPtr)
        If DX_Device IsNot Nothing Then
            DX_Device.Dispose()
            DX_Device = Nothing
        End If

        If Present_Params Is Nothing Then
            Present_Params = New PresentParameters
        End If
        With Present_Params

            .Windowed = True

            If DEVICE_BackBufferWidth = 0 Then
                .BackBufferWidth = Screen.PrimaryScreen.Bounds.Width
                .BackBufferHeight = Screen.PrimaryScreen.Bounds.Height
            Else
                .BackBufferWidth = DEVICE_BackBufferWidth
                .BackBufferHeight = DEVICE_BackBufferHeight
            End If

            'If DEVICE_HardwareFullScreen Then
            '    .BackBufferWidth = Screen.PrimaryScreen.Bounds.Width
            '    .BackBufferHeight = Screen.PrimaryScreen.Bounds.Height
            '    DEVICE_BackBufferWidth = .BackBufferWidth
            '    DEVICE_BackBufferHeight = .BackBufferHeight
            '    .BackBufferFormat = Format.X8R8G8B8
            '    .Windowed = False
            'Else
            '    ' use the system format X8R8G8B8 ( or do not set it ) to maximize speed on secondary monitors
            '    '.BackBufferFormat = Format.R5G6B5
            '    '.BackBufferFormat = Format.X8R8G8B8 'Format.X8R8G8B8 'Format.R5G6B5
            'End If

            .DeviceWindowHandle = WindowHandle

            .BackBufferCount = 1


            .AutoDepthStencilFormat = DepthFormat.D24S8 'DepthFormat.D16 ' DepthFormat.D24S8 ' DepthFormat.D24X8
            .EnableAutoDepthStencil = True

            .MultiSample = CType(DEVICE_Multisamples, MultiSampleType)

            .SwapEffect = SwapEffect.Discard
            .PresentationInterval = PresentInterval.Immediate

        End With

        If DEVICE_HardwareVertexProcessing Then
            Try
                DX_Device = New Device(0, DeviceType.Hardware, WindowHandle, CreateFlags.HardwareVertexProcessing, Present_Params)
            Catch
            End Try
        End If

        If DX_Device Is Nothing And DEVICE_HardwareVertexProcessing Then
            Try
                Present_Params.MultiSample = MultiSampleType.None
                DX_Device = New Device(0, DeviceType.Hardware, WindowHandle, CreateFlags.HardwareVertexProcessing, Present_Params)
            Catch
            End Try
        End If

        If DX_Device Is Nothing Then
            Try
                DEVICE_HardwareVertexProcessing = False
                Present_Params.MultiSample = CType(DEVICE_Multisamples, MultiSampleType)
                DX_Device = New Device(0, DeviceType.Hardware, WindowHandle, CreateFlags.SoftwareVertexProcessing, Present_Params)
            Catch
            End Try
        End If

        If DX_Device Is Nothing Then
            Try
                Present_Params.MultiSample = MultiSampleType.None
                DX_Device = New Device(0, DeviceType.Hardware, WindowHandle, CreateFlags.SoftwareVertexProcessing, Present_Params)
            Catch
            End Try
        End If

        If DX_Device Is Nothing Then
            Try
                DX_Device = New Device(0, DeviceType.Software, WindowHandle, CreateFlags.SoftwareVertexProcessing, Present_Params)
            Catch
                MsgBox("Change PC che l'e' mei.", MsgBoxStyle.Information)
                End
            End Try
        End If

        DEVICE_Multisamples = Present_Params.MultiSample
    End Sub





    ' ===============================================================================================
    '   MOUSE
    ' ===============================================================================================
    Private StartCursorPos As Point = New Point(-1, -1)
    Friend Sub DX_MouseDown(ByVal e As MouseEventArgs)
        StartCursorPos.X = e.X
        StartCursorPos.Y = e.Y
        If e.Button = MouseButtons.Left Then
            DirectX_MoveObjectStart(e.X, e.Y)
        End If
        If e.Button = MouseButtons.Right Then
            DX_ShootObject(e.X, e.Y)
        End If
    End Sub

    Friend Sub DX_MouseMove(ByVal e As MouseEventArgs)
        If StartCursorPos.X < 0 Then Exit Sub
        If e.Button = MouseButtons.Left Then
            If DX_EditingObject Then
                DX_MoveObject(e.X, e.Y)
            Else
                DX_MoveCamera(e.X - StartCursorPos.X, e.Y - StartCursorPos.Y)
                'EventsAreEnabled = False
                'Cursor.Position = PictureBox1.PointToScreen(StartCursorPos)
                'EventsAreEnabled = True
            End If
        End If
        StartCursorPos.X = e.X
        StartCursorPos.Y = e.Y
    End Sub
    Friend Sub DX_MouseUp()
        If DX_EditingObject Then
            Dim obj As Object_3D = Objects_3D(DX_EditingObject_DxObjectIndex)
            If obj.ObjType = ObjTypes.CLOTH Or obj.ObjType = ObjTypes.SOFT_OBJECT Or obj.ObjType = ObjTypes.ROPE Then
                Bullet_SoftObj_SetDamping(obj.BulletIndex, oldDamping)
            Else
                Bullet_Object_SetDamping(obj.BulletIndex, oldLinearDamping, oldAngularDamping)
            End If
            ' ---------- restore normal SleepingThresholds used to substitute the ( not working ) SetActivationState
            Bullet_Object_SetSleepingThresholds(obj.BulletIndex, 0.5, 0.5)
            Bullet_Object_SetActivationState(obj.BulletIndex, ActivationState.ACTIVE_TAG)
            ' ---------------------------------------------------------------------------------------------------
            DX_EditingObject = False
            DX_EditingObject_DxObjectIndex = -1
        End If
        StartCursorPos.X = -1
    End Sub




    ' ===============================================================================================
    '   MOVE OBJECTS
    ' ===============================================================================================

    Friend DX_EditingObject As Boolean = False
    Friend DX_EditingObject_DxObjectIndex As Int32 = -1
    Private MouseX As Int32
    Private Mousey As Int32
    Private Node As Int32
    Private oldDamping As Single
    Private oldLinearDamping As Single
    Private oldAngularDamping As Single

    Private Sub TransformXyToViewportSize(ByRef x As Int32, ByRef y As Int32)
        x = (x * DX_Device.Viewport.Width) \ Form1.DisplayRectangle.Width
        y = (y * DX_Device.Viewport.Height) \ Form1.DisplayRectangle.Height
    End Sub

    Friend Sub DirectX_MoveObjectStart(ByVal x As Int32, ByVal y As Int32)
        Dim wx As Int32 = Form1.DisplayRectangle.Width
        Dim wy As Int32 = Form1.DisplayRectangle.Height

        TransformXyToViewportSize(x, y)

        DX_UpdateCameraParams()
        Dim ExcludedObject As Int32 = -1
        If ControlType = ControlTypes.MoveVehicleWithInternalCamera Then ExcludedObject = VehicleActive
        Pick_FromViewport(x, y, ExcludedObject)
        DX_EditingObject_DxObjectIndex = Pick_GetPickedObjIndex()
        If DX_EditingObject_DxObjectIndex >= 0 Then

            Dim obj As Object_3D = Objects_3D(DX_EditingObject_DxObjectIndex)

            If Form2.Visible Then
                Form2.Label_Debug1.Text = "    DX_Object = " & DX_EditingObject_DxObjectIndex.ToString & vbCr
                Form2.Label_Debug1.Text &= "Bullet object = " & obj.BulletIndex.ToString & vbCr
            End If

            ' ------------------------------- SetSleepingThresholds substitutes the ( not working ) SetActivationState
            Bullet_Object_SetSleepingThresholds(obj.BulletIndex, 0, 0)
            Bullet_Object_SetActivationState(obj.BulletIndex, ActivationState.ACTIVE_TAG)
            ' --------------------------------------------------------------------------------------------------------

            'If obj.ObjType = ObjTypes.CLOTH Then
            '    Bullet_SoftObj_Cut(obj.BulletIndex, ViewportPick_GetPickedPoint_Local(), 1)
            '    ' warning - obj is read only - can not set obj props !
            '    obj.MeshX.msh = Nothing
            '    obj.MeshX.msh = Bullet_SoftObj_GetMesh(obj)
            '    Return
            'End If

            If obj.IsMovingObject Then
                DX_EditingObject = True
                MouseX = x
                Mousey = y
                oldLinearDamping = Bullet_Object_GetLinearDamping(obj.BulletIndex)
                oldAngularDamping = Bullet_Object_GetAngularDamping(obj.BulletIndex)
            Else
                If obj.ObjType = ObjTypes.CLOTH Or obj.ObjType = ObjTypes.SOFT_OBJECT Then
                    DX_EditingObject = True
                    MouseX = x
                    Mousey = y
                    Node = ViewportPick_GetVertexId()
                    'Form1.Label99.Text = Node.ToString
                    oldDamping = Bullet_SoftObj_GetDamping(obj.BulletIndex)
                End If
                If obj.ObjType = ObjTypes.ROPE Then
                    DX_EditingObject = True
                    MouseX = x
                    Mousey = y
                    Node = ViewportPick_GetVertexId() \ (obj.RopeFaces + 1)
                    'Form1.Label99.Text = Node.ToString
                    oldDamping = Bullet_SoftObj_GetDamping(obj.BulletIndex)
                End If
                If Form2.Visible Then
                    Form2.Label_Debug1.Text &= "  Picked node = " & Node.ToString
                End If
            End If
        End If
    End Sub
    Friend Sub DX_MoveObject(ByVal x As Int32, ByVal y As Int32)
        TransformXyToViewportSize(x, y)
        MouseX = x
        Mousey = y
    End Sub


    Private Sub MoveObject()
        Dim obj As Object_3D = Objects_3D(DX_EditingObject_DxObjectIndex)

        Select Case obj.ObjType

            Case ObjTypes.SOFT_OBJECT, ObjTypes.CLOTH, ObjTypes.ROPE
                'Try
                Dim Pos3D As Vector3 = obj.SoftObjNodes(Node)
                Dim PosScr As Point = ObjectPosToScreenPos(Pos3D)
                Dim dx As Int32 = MouseX - PosScr.X
                Dim dy As Int32 = Mousey - PosScr.Y
                Dim push As Vector3
                ' ---------------------------------------------------------------
                Dim kdist As Single = (Pos3D - Camera.Pos).Length
                If kdist < 2 Then kdist = 0
                'If kdist > 100 Then kdist = 100
                kdist *= 10
                ' ---------------------------------------------------------------
                If Key(Keys.ShiftKey) Then
                    push = VEC3(dx, -dy, kdist) * 0.3
                ElseIf Key(Keys.ControlKey) Then
                    push = VEC3(dx, -dy, -kdist) * 0.3
                Else
                    push = VEC3(dx, -dy, 0) * 0.005 * kdist
                End If
                'push.X = 0.001F * Math.Sign(push.X) * (push.X * push.X)
                'push.Y = 0.001F * Math.Sign(push.Y) * (push.Y * push.Y)
                'push.Z = 0.001F * Math.Sign(push.Z) * (push.Z * push.Z)
                push.TransformNormal(Matrix.Invert(mtx_CameraView))
                push = LimitVector3(push, 100)

                Bullet_SoftObj_SetDamping(obj.BulletIndex, 0.07)

                Dim objMass As Single = Bullet_Object_GetMass(obj.BulletIndex)
                Bullet_SoftObj_ApplyForce(obj.BulletIndex, push * objMass, Node)
                'Catch
                ' App_StartScene()
                'End Try

            Case Else
                Dim Pos3D As Vector3 = Matrix_GetVectorPos(obj.MeshX.mtx)
                Dim PosScr As Point = ObjectPosToScreenPos(Pos3D)
                Dim dx As Int32 = MouseX - PosScr.X
                Dim dy As Int32 = Mousey - PosScr.Y
                Dim push As Vector3
                ' ---------------------------------------------------------------
                Dim kdist As Single = (Pos3D - Camera.Pos).Length
                If kdist < 2 Then kdist = 0
                kdist *= 10
                ' ---------------------------------------------------------------
                If Key(Keys.ShiftKey) Then
                    push = VEC3(dx, -dy, kdist) * 0.07
                ElseIf Key(Keys.ControlKey) Then
                    push = VEC3(dx, -dy, -kdist) * 0.07
                Else
                    push = VEC3(dx, -dy, 0) * 0.07
                End If
                push.TransformNormal(Matrix.Invert(mtx_CameraView))
                push = LimitVector3(push, 10)
                ' ------------------------------------------------------------------------------------
                Bullet_Object_SetActivationState(obj.BulletIndex, ActivationState.ACTIVE_TAG)
                Dim objMass As Single = Bullet_Object_GetMass(obj.BulletIndex)
                ' ------------------------------------------------------------------------------------
                Bullet_Object_SetDamping(obj.BulletIndex, 0.999, 0.99)
                Bullet_Object_ApplyDamping(obj.BulletIndex, 0.1)
                Bullet_Object_ApplyCentralImpulse(obj.BulletIndex, push * objMass)

        End Select
    End Sub


    Private Function LimitVector3(ByVal vector As Vector3, ByVal max As Single) As Vector3
        If vector.Length > max Then
            Return Vector3.Normalize(vector) * max
        Else
            Return vector
        End If
    End Function



    Private Function ObjectPosToScreenPos(ByVal ObjPos As Vector3) As Point
        Dim p As Vector3 = Vector3.Project(ObjPos, _
                                           DX_Device.Viewport, _
                                           mtx_CameraProj, _
                                           mtx_CameraView, _
                                           Matrix.Identity)
        If p.X > 9000 Then p.X = 9000
        If p.X < -9000 Then p.X = -9000
        If p.Y > 9000 Then p.Y = 9000
        If p.X < -9000 Then p.X = -9000
        Return New Point(CInt(p.X), CInt(p.Y))
    End Function



    ' ===============================================================================================
    '   SHOOT SPHERES
    ' ===============================================================================================

    Private ShootIndex As Int32 = -1
    Friend Sub DX_ShootObject(ByVal x As Int32, ByVal y As Int32)
        If ShootSpheres_Last < ShootSpheres_First Then Return
        ShootIndex += 1
        If ShootIndex < ShootSpheres_First Or ShootIndex > ShootSpheres_Last Then ShootIndex = ShootSpheres_First
        DX_UpdateCameraParams()
        Dim destPosition As Vector3 = ViewportPick_GetRayTo(x, y)
        Bullet_Object_Shoot(Objects_3D(ShootIndex).BulletIndex, Camera.Pos + Camera.Dir * 1, destPosition, f_ShootSpeed)
        Objects_3D(ShootIndex).Shooted = True
    End Sub





    ' ===============================================================================================
    '   MOVEMENTS
    ' ===============================================================================================


    Friend Sub DX_Step()
        If DX_EditingObject Then MoveObject()
        If LookAtObjectIndex >= 0 Then LookAtObject()
        MoveByKeys()
        CameraFollowVehicle()
        CameraInternalToVehicle()
        LimitCameraMovements()
        DX_UpdateCameraParams()

        mtx_CameraProj = Matrix.PerspectiveFovLH(50 / Camera.Lens, f_ViewAspect, f_NearPlane, f_FarPlane)
        SetSoundListenerWithMatrix(mtx_CameraView)
        DX_DrawAll()
    End Sub



    Private Sub MoveByKeys()
        Select Case ControlType
            Case ControlTypes.MoveCamera
                MoveCamera()
            Case ControlTypes.MoveVehicle
                App_MoveVehiclesWithKeys()
                ' CameraFollowVehicle()
            Case ControlTypes.MoveVehicleWithInternalCamera
                App_MoveVehiclesWithKeys()
                'CameraInternalToVehicle()
            Case ControlTypes.AnimateVehicles
                MoveCamera()
                App_AnimateVehicles()
        End Select
    End Sub

    Private Sub MoveCamera()
        If Not Form1.Focused Then Exit Sub

        Dim linSpeed As Single = 0.4
        Dim rotSpeed As Single = 0.05
        If My.Computer.Keyboard.CtrlKeyDown Then
            linSpeed *= 6
            rotSpeed *= 2
        End If

        If Key(Keys.Up) Then
            If Not Key(Keys.ShiftKey) Then
                mtx_CameraView *= Matrix.Translation(VEC3(0, 0, -linSpeed * 1.5F))
            Else
                mtx_CameraView *= Matrix.RotationAxis(VEC3(1, 0, 0), rotSpeed)
            End If
        End If
        If Key(Keys.Down) Then
            If Not Key(Keys.ShiftKey) Then
                mtx_CameraView *= Matrix.Translation(VEC3(0, 0, linSpeed * 1.5F))
            Else
                mtx_CameraView *= Matrix.RotationAxis(VEC3(1, 0, 0), -rotSpeed)
            End If
        End If
        If Key(Keys.Left) Then
            If Not Key(Keys.ShiftKey) Then
                mtx_CameraView *= Matrix.Translation(VEC3(linSpeed, 0, 0))
            Else
                mtx_CameraView *= Matrix.RotationAxis(Matrix_GetVectorUp(mtx_CameraView), rotSpeed)
            End If
        End If
        If Key(Keys.Right) Then
            If Not Key(Keys.ShiftKey) Then
                mtx_CameraView *= Matrix.Translation(VEC3(-linSpeed, 0, 0))
            Else
                mtx_CameraView *= Matrix.RotationAxis(Matrix_GetVectorUp(mtx_CameraView), -rotSpeed)
            End If
        End If
        If Key(Keys.PageUp) Then
            mtx_CameraView *= Matrix.Translation(VEC3(0, -linSpeed, 0))
        End If
        If Key(Keys.PageDown) Then
            mtx_CameraView *= Matrix.Translation(VEC3(0, linSpeed, 0))
        End If
        '
        ' ---------------------------------------------------------------- Move camera with Slots
        Dim push As Single
        Dim dir As Single
        If SlotUp >= 0 Then
            push = Slots.ReadSlot_NoNan_0_1000(SlotUp)
            push = -0.1F * (push - 500)
        End If
        If SlotDir >= 0 Then
            dir = Slots.ReadSlot_NoNan_0_1000(SlotDir)
            dir = 0.1F * (dir - 500)
        End If
        If push <> 0 Or dir <> 0 Then
            DX_MoveCamera(CInt(dir), CInt(push))
        End If
    End Sub


    ' ===============================================================================================
    '   MOVE CAMERA
    ' ===============================================================================================

    Private Sub DX_MoveCamera(ByVal dx As Int32, ByVal dy As Int32)
        If UseCameraObject Then
            With Objects_3D(CameraIndex)
                .MeshX.mtx *= Matrix.RotationAxis(VEC3(1, 0, 0), -dy * 0.004F)
                .MeshX.mtx *= Matrix.RotationAxis(Matrix_GetVectorUp(mtx_CameraView), -dx * 0.004F)
                mtx_CameraView = .MeshX.mtx
                Bullet_Object_SetMatrix(.BulletIndex, .MeshX.mtx)
            End With
        Else
            mtx_CameraView *= Matrix.RotationAxis(VEC3(1, 0, 0), -dy * 0.004F)
            mtx_CameraView *= Matrix.RotationAxis(Matrix_GetVectorUp(mtx_CameraView), -dx * 0.004F)
        End If
    End Sub

    Friend Sub DX_MouseWheel(ByVal delta As Int32)
        If My.Computer.Keyboard.CtrlKeyDown Then
            delta *= 10
        End If
        If UseCameraObject Then
            With Objects_3D(CameraIndex).MeshX
                .mtx *= Matrix.Translation(VEC3(0, 0, -delta * 0.02F))
                mtx_CameraView = .mtx
            End With
        Else
            mtx_CameraView *= Matrix.Translation(VEC3(0, 0, -delta * 0.02F))
        End If
    End Sub

    Friend Sub DX_ChangeLens(ByVal delta As Int32)
        Camera.Lens *= (1 + delta / 1000.0F)
        If Camera.Lens < 25 Then Camera.Lens = 25
        If Camera.Lens > 100 Then Camera.Lens = 100
    End Sub


    Friend Sub DX_ChangeControlType()
        If VehicleActive >= 0 Then
            ControlType = CType((ControlType + 1) Mod 3, ControlTypes)
        Else
            ControlType = ControlTypes.MoveCamera
            LookAtObjectIndex = -1
        End If
        If ControlType = ControlTypes.MoveCamera Then
            App_Vehicle_StopMotors()
        End If
        If VehicleActive >= 0 And ControlType = ControlTypes.MoveCamera Then
            DX_MouseWheel(-300)
        End If
    End Sub



    Private Sub CameraFollowVehicle()
        If VehicleActive < 0 Then Return
        If ControlType <> ControlTypes.MoveVehicle Then Return

        Dim obj As Object_3D = Objects_3D(VehicleActive)
        Dim carspeed As Single = obj.LocalSpeed.Z

        Dim cm As Matrix = obj.MeshX.mtx
        Dim pos As Vector3 = Matrix_GetVectorPos(cm)
        Dim dir As Vector3 = Matrix_GetVectorDir(cm)
        dir.Y = 0

        Dim pos2 As Vector3 = pos - dir * 10
        Dim delta As Vector3 = pos2 - Camera.Pos

        Dim maxdist As Single = 4
        Dim followspeed As Single = 0.005F + Math.Abs(carspeed / 200.0F)

        Dim speed As Single = delta.Length - maxdist
        If speed > 1 Then speed = 1
        If speed < 0 Then speed = 0
        Camera.Pos += delta * speed * followspeed
        Camera.Pos.Y = pos.Y + VehicleCameraHeight
        mtx_CameraView = Matrix.LookAtLH(Camera.Pos, pos, VEC3(0, 1, 0))
    End Sub


    Private Sub CameraInternalToVehicle()
        If VehicleActive < 0 Then Exit Sub
        If ControlType <> ControlTypes.MoveVehicleWithInternalCamera Then Return
        Dim obj As Object_3D = Objects_3D(VehicleActive)
        Dim m As Matrix = obj.MeshX.mtx
        Dim pos As Vector3 = Matrix_GetVectorPos(m)
        Dim right As Vector3 = Matrix_GetVectorRight(m)
        Dim up As Vector3 = Matrix_GetVectorUp(m)
        Dim dir As Vector3 = Matrix_GetVectorDir(m)
        'pos += -0.3F * Vector3.Normalize(right)
        'pos += 1.6F * Vector3.Normalize(up)
        'pos += 0.75F * Vector3.Normalize(dir)
        pos += -0.3F * right
        pos += 1.0F * up
        pos += 1.05F * dir
        mtx_CameraView = Matrix.LookAtLH(pos, pos + dir, up)
    End Sub


    Private Sub LookAtObject()
        Dim obj As Object_3D = Objects_3D(LookAtObjectIndex)
        Dim pos As Vector3 = Bullet_Object_GetPosition(obj.BulletIndex)
        DX_UpdateCameraParams()
        Dim m As Matrix = Matrix.LookAtLH(Camera.Pos, pos, VEC3(0, 1, 0))
        If Not Single.IsNaN(pos.X) Then
            mtx_CameraView = m
        End If
    End Sub




    ' ===============================================================================================
    '   CAMERA
    ' ===============================================================================================

    Private UseCameraObject As Boolean = False

    Friend Sub DX_SetCamera()
        Camera.Lens = 50
        Camera.Pos = VEC3(0, 3, -13)
        Camera.At = VEC3(0, 3, 0)
        Camera.Up = VEC3(0, 1, 0)
        mtx_CameraView = Matrix.LookAtLH(Camera.Pos, Camera.At, Camera.Up)
    End Sub

    Friend Sub DX_SetCamera(ByVal Pos As Vector3, ByVal At As Vector3)
        Camera.Pos = Pos
        Camera.At = At
        Camera.Up = VEC3(0, 1, 0)
        mtx_CameraView = Matrix.LookAtLH(Camera.Pos, Camera.At, Camera.Up)
    End Sub

    Friend Sub DX_SetCamera(ByVal CameraY As Single)
        Camera.Pos.Y = CameraY
        mtx_CameraView = Matrix.LookAtLH(Camera.Pos, Camera.At, Camera.Up)
    End Sub


    Friend Sub DX_UpdateCameraParams()
        Dim mvi As Matrix = Matrix.Invert(mtx_CameraView)
        Camera.Pos = Matrix_GetVectorPos(mvi)
        Camera.Dir = Matrix_GetVectorDir(mvi)
        Camera.Up = Matrix_GetVectorUp(mvi)
        Camera.Right = Matrix_GetVectorRight(mvi)
        Camera.At = Camera.Pos + Camera.Dir
    End Sub

    Private Sub LimitCameraMovements()
        If ControlType = ControlTypes.MoveVehicleWithInternalCamera Then Return
        Dim mvi As Matrix = Matrix.Invert(mtx_CameraView)
        If mvi.M42 < 1.8 Then
            mvi.M42 = 1.8
            mtx_CameraView = Matrix.Invert(mvi)
        End If
    End Sub









End Module












