﻿Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

Module Functions


    ' ##############################################################################################
    ' ########
    ' ########      VAR PTR
    ' ########
    ' ##############################################################################################
    ' ##############################################################################################
    ' ##############################################################################################

    Friend Function VarPtr(ByVal e As Object) As IntPtr
        Dim GC As System.Runtime.InteropServices.GCHandle = _
        System.Runtime.InteropServices.GCHandle.Alloc(e, System.Runtime.InteropServices.GCHandleType.Pinned)
        Dim GC2 As Int32 = GC.AddrOfPinnedObject.ToInt32
        GC.Free()
        Return CType(GC2, IntPtr)
    End Function

    Friend Function VarPtr_Int(ByVal e As Object) As Int32
        Dim GC As System.Runtime.InteropServices.GCHandle = _
        System.Runtime.InteropServices.GCHandle.Alloc(e, System.Runtime.InteropServices.GCHandleType.Pinned)
        Dim GC2 As Int32 = GC.AddrOfPinnedObject.ToInt32
        GC.Free()
        Return GC2
    End Function




    ' ##############################################################################################
    ' ########
    ' ########      ASYNC KEY STATE
    ' ########
    ' ##############################################################################################
    ' ##############################################################################################
    ' ##############################################################################################

    Private Declare Function GetAsyncKeyState Lib "user32" (ByVal vKey As Int32) As Int16

    Private Const PRESSED_NOW As Int32 = &H8000
    Private Const PRESSED_AFTER_PREVIOUS_CALL As Int32 = &H1

    Friend Function Key(ByVal k As Int32) As Boolean
        Key = (GetAsyncKeyState(k) And PRESSED_NOW) <> 0
    End Function

    Friend Function KeyPrevious(ByVal k As Int32) As Boolean
        KeyPrevious = (GetAsyncKeyState(k) And PRESSED_AFTER_PREVIOUS_CALL) <> 0
    End Function



    ' ##############################################################################################
    ' ########
    ' ########        RENDER TARGETS
    ' ########
    ' ##############################################################################################
    ' ##############################################################################################
    ' ##############################################################################################

    Friend Structure RenderTarget
        Dim texture As Texture
        Dim surface0 As Surface
        Dim stencil As Surface
        Dim bbuffer_Saved As Surface
        Dim stencil_Saved As Surface

        Sub Init(ByVal w As Int32, _
                 ByVal h As Int32, _
                 ByVal fmt As Format)
            '
            ' -------------------------------------------------------------------- Create the shadow map texture
            ' Formats - R32F(SMAP) - X8R8G8B8(FastVr)
            '
            texture = New Texture(DX_Device, w, h, 1, Usage.RenderTarget, fmt, Pool.Default)
            '
            surface0 = texture.GetSurfaceLevel(0)
            '
            ' -------------------------------------------------------------------- Create the depth-stencil buffer
            ' format = D32 or D16
            '   or "Present_Params.AutoDepthStencilFormat = D24S8" 
            ' discard = False(FastVr) or True(bullet wrapper shadows)
            '
            stencil = DX_Device.CreateDepthStencilSurface(w, h, DepthFormat.D16, MultiSampleType.None, 0, True)
            '
            ' -------------------------------------------------------------------- init all
            bbuffer_Saved = Nothing
            stencil_Saved = Nothing
        End Sub

        Friend Sub Destroy()
            If surface0 IsNot Nothing Then
                surface0.Dispose()
                surface0 = Nothing
            End If
            If texture IsNot Nothing Then
                texture.Dispose()
                texture = Nothing
            End If
            If stencil IsNot Nothing Then
                stencil.Dispose()
                stencil = Nothing
            End If
        End Sub

        Friend Sub SetRenderTargetToDevice()
            bbuffer_Saved = DX_Device.GetRenderTarget(0)
            stencil_Saved = DX_Device.DepthStencilSurface
            DX_Device.SetRenderTarget(0, surface0)
            DX_Device.DepthStencilSurface = stencil
        End Sub

        Friend Sub RestoreOldRenderTarget()
            If bbuffer_Saved IsNot Nothing Then
                DX_Device.SetRenderTarget(0, bbuffer_Saved)
                bbuffer_Saved.Dispose()
                bbuffer_Saved = Nothing
            End If
            If stencil_Saved IsNot Nothing Then
                DX_Device.DepthStencilSurface = stencil_Saved
                stencil_Saved.Dispose()
                stencil_Saved = Nothing
            End If
        End Sub

    End Structure














    ' ===============================================================================================
    '   UTILITY FUNCTIONS
    ' ===============================================================================================

    Friend Function Matrix_Build(ByVal m11 As Single, ByVal m12 As Single, ByVal m13 As Single, ByVal m14 As Single, _
                                 ByVal m21 As Single, ByVal m22 As Single, ByVal m23 As Single, ByVal m24 As Single, _
                                 ByVal m31 As Single, ByVal m32 As Single, ByVal m33 As Single, ByVal m34 As Single, _
                                 ByVal m41 As Single, ByVal m42 As Single, ByVal m43 As Single, ByVal m44 As Single) As Matrix
        Matrix_Build.M11 = m11
        Matrix_Build.M12 = m12
        Matrix_Build.M13 = m13
        Matrix_Build.M14 = m14

        Matrix_Build.M21 = m21
        Matrix_Build.M22 = m22
        Matrix_Build.M23 = m23
        Matrix_Build.M24 = m24

        Matrix_Build.M31 = m31
        Matrix_Build.M32 = m32
        Matrix_Build.M33 = m33
        Matrix_Build.M34 = m34

        Matrix_Build.M41 = m41
        Matrix_Build.M42 = m42
        Matrix_Build.M43 = m43
        Matrix_Build.M44 = m44
    End Function

    Friend Function Matrix_GetVectorRight(ByRef m As Matrix) As Vector3
        Matrix_GetVectorRight = VEC3(m.M11, m.M12, m.M13)
    End Function

    Friend Function Matrix_GetVectorUp(ByRef m As Matrix) As Vector3
        Matrix_GetVectorUp = VEC3(m.M21, m.M22, m.M23)
    End Function

    Friend Function Matrix_GetVectorDir(ByRef m As Matrix) As Vector3
        Matrix_GetVectorDir = VEC3(m.M31, m.M32, m.M33)
    End Function

    Friend Function Matrix_GetVectorPos(ByRef m As Matrix) As Vector3
        Matrix_GetVectorPos = VEC3(m.M41, m.M42, m.M43)
    End Function

    Friend Function Matrix_GetScale(ByVal m As Matrix) As Vector3
        Matrix_GetScale.X = VEC3(m.M11, m.M12, m.M13).Length
        Matrix_GetScale.Y = VEC3(m.M21, m.M22, m.M23).Length
        Matrix_GetScale.Z = VEC3(m.M31, m.M32, m.M33).Length
    End Function

    Friend Sub DX_ComputeBoundingBoxFromMesh(ByVal mesh As Mesh, ByRef bbmin As Vector3, ByRef bbmax As Vector3)
        If mesh Is Nothing Then Return
        Dim vertexdata As GraphicsStream
        vertexdata = mesh.VertexBuffer.Lock(0, 0, LockFlags.None)
        Geometry.ComputeBoundingBox(vertexdata, _
                                    mesh.NumberVertices, _
                                    mesh.VertexFormat, _
                                    bbmin, _
                                    bbmax)
        mesh.VertexBuffer.Unlock()
    End Sub

    Friend Sub DX_ComputeBoundingSphereFromMesh(ByVal mesh As Mesh, ByRef SphereRadius As Single, ByRef SphereCenter As Vector3)
        Dim vertexdata As GraphicsStream
        vertexdata = mesh.VertexBuffer.Lock(0, 0, LockFlags.None)
        SphereRadius = Geometry.ComputeBoundingSphere(vertexdata, _
                                                      mesh.NumberVertices, _
                                                      mesh.VertexFormat, _
                                                      SphereCenter)
        mesh.VertexBuffer.Unlock()
    End Sub

    Friend Function VEC3(ByVal x As Single, ByVal y As Single, ByVal z As Single) As Vector3
        Return New Vector3(x, y, z)
    End Function

    Friend Function VEC4(ByVal x As Single, ByVal y As Single, ByVal z As Single, ByVal w As Single) As Vector4
        Return New Vector4(x, y, z, w)
    End Function


    Friend Function CreateDefaultTexture() As Texture
        CreateDefaultTexture = New Texture(DX_Device, _
                                           32, 32, 1, _
                                           Usage.None, _
                                           Format.X8R8G8B8, _
                                           Pool.Managed)

        Dim s As Surface = CreateDefaultTexture.GetSurfaceLevel(0)
        Dim g As System.Drawing.Graphics = s.GetGraphics()
        g.Clear(Color.FromArgb(&HFF808080))
        s.ReleaseGraphics()
    End Function


    Friend Function CreateTexture(ByVal w As Int32, _
                                  ByVal h As Int32, _
                                  ByVal FillColor As Color, _
                                  Optional ByVal _Usage As Usage = Usage.None, _
                                  Optional ByVal _Format As Format = Format.A8R8G8B8, _
                                  Optional ByVal _Pool As Pool = Pool.Managed) As Texture

        CreateTexture = New Texture(DX_Device, w, h, 1, _Usage, _Format, _Pool)

        Dim s As Surface = CreateTexture.GetSurfaceLevel(0)
        Dim g As System.Drawing.Graphics = s.GetGraphics()
        g.Clear(FillColor)
        s.ReleaseGraphics()
    End Function




    ' ##############################################################################################
    ' ########
    ' ########        BOUNDING BOX
    ' ########
    ' ##############################################################################################
    ' ##############################################################################################
    ' ##############################################################################################

    Friend Structure BoundingBox
        Dim minPt As Vector3
        Dim maxPt As Vector3
        Sub Init(ByVal points() As Vector3)
            minPt = New Vector3(3.3E+33F, 3.3E+33F, 3.3E+33F)
            maxPt = New Vector3(-3.3E+33F, -3.3E+33F, -3.3E+33F)
            For i As Int32 = 0 To points.Length - 1
                Merge(points(i))
            Next
        End Sub
        Sub Init(ByVal bbox As BoundingBox)
            minPt = bbox.minPt
            maxPt = bbox.maxPt
        End Sub
        Sub Init(ByVal bbox() As BoundingBox)
            minPt = New Vector3(3.3E+33F, 3.3E+33F, 3.3E+33F)
            maxPt = New Vector3(-3.3E+33F, -3.3E+33F, -3.3E+33F)
            For i As Int32 = 0 To bbox.Length - 1
                Merge(bbox(i).maxPt)
                Merge(bbox(i).minPt)
            Next
        End Sub
        Sub Init(ByVal _minPt As Vector3, ByVal _maxPt As Vector3)
            minPt = _minPt
            maxPt = _maxPt
        End Sub
        Sub Init()
            minPt = New Vector3(3.3E+33F, 3.3E+33F, 3.3E+33F)
            maxPt = New Vector3(-3.3E+33F, -3.3E+33F, -3.3E+33F)
        End Sub
        Function Centroid() As Vector3
            Return 0.5F * (minPt + maxPt)
        End Function
        Function Size() As Single
            Return (maxPt - minPt).Length
        End Function
        Function Point_P(ByVal i As Int32) As Vector3
            Select Case i
                Case 0 : Return VEC3(minPt.X, minPt.Y, minPt.Z)
                Case 1 : Return VEC3(maxPt.X, minPt.Y, minPt.Z)
                Case 2 : Return VEC3(minPt.X, maxPt.Y, minPt.Z)
                Case 3 : Return VEC3(maxPt.X, maxPt.Y, minPt.Z)
                Case 4 : Return VEC3(minPt.X, minPt.Y, maxPt.Z)
                Case 5 : Return VEC3(maxPt.X, minPt.Y, maxPt.Z)
                Case 6 : Return VEC3(minPt.X, maxPt.Y, maxPt.Z)
                Case 7 : Return VEC3(maxPt.X, maxPt.Y, maxPt.Z)
            End Select
        End Function
        Function Point_N(ByVal i As Int32) As Vector3
            Select Case i
                Case 0 : Return VEC3(maxPt.X, maxPt.Y, maxPt.Z)
                Case 1 : Return VEC3(minPt.X, maxPt.Y, maxPt.Z)
                Case 2 : Return VEC3(maxPt.X, minPt.Y, maxPt.Z)
                Case 3 : Return VEC3(minPt.X, minPt.Y, maxPt.Z)
                Case 4 : Return VEC3(maxPt.X, maxPt.Y, minPt.Z)
                Case 5 : Return VEC3(minPt.X, maxPt.Y, minPt.Z)
                Case 6 : Return VEC3(maxPt.X, minPt.Y, minPt.Z)
                Case 7 : Return VEC3(minPt.X, minPt.Y, minPt.Z)
            End Select
        End Function
        Sub Merge(ByVal vec As Vector3)
            minPt.X = Math.Min(minPt.X, vec.X)
            minPt.Y = Math.Min(minPt.Y, vec.Y)
            minPt.Z = Math.Min(minPt.Z, vec.Z)
            maxPt.X = Math.Max(maxPt.X, vec.X)
            maxPt.Y = Math.Max(maxPt.Y, vec.Y)
            maxPt.Z = Math.Max(maxPt.Z, vec.Z)
        End Sub
        'Function Intersect(ByRef hitDist As Single, ByVal origPt As Vector3, ByVal dir As Vector3) As Boolean
        '    Dim sides(5) As Plane
        '    sides(0) = New Plane(1, 0, 0, -minPt.X)
        '    sides(1) = New Plane(-1, 0, 0, maxPt.X)
        '    sides(2) = New Plane(0, 1, 0, -minPt.Y)
        '    sides(3) = New Plane(0, -1, 0, maxPt.Y)
        '    sides(4) = New Plane(0, 0, 1, -minPt.Z)
        '    sides(5) = New Plane(0, 0, -1, maxPt.Z)

        '    hitDist = 0
        '    Dim hitPt As Vector3 = origPt
        '    Dim inside As Boolean = False
        '    Dim cosTheta As Single
        '    Dim dist As Single

        '    For i As Int32 = 0 To 5
        '        cosTheta = Plane.DotNormal(sides(i), dir)
        '        dist = Plane_DotCoordinate(sides(i), origPt)
        '        '  ------------------------------------------- nearly intersecting, just punt and call it an intersection
        '        If ALMOST_ZERO(dist) Then Return True
        '        '  ------------------------------------------- skip nearly (&actually) parallel rays
        '        If ALMOST_ZERO(cosTheta) Then Continue For
        '        '  ------------------------------------------- only interested in intersections along the ray, not before it.
        '        hitDist = -dist / cosTheta
        '        If hitDist < 0 Then Continue For

        '        hitPt = hitDist * dir + origPt

        '        inside = True

        '        If inside Then
        '            For j As Int32 = 0 To 5
        '                If j = i Then Continue For
        '                inside = (Plane_DotCoordinate(sides(j), hitPt) + 0.00015) >= 0
        '                If Not inside Then Exit For
        '            Next
        '        End If

        '        If inside Then Exit For
        '    Next

        '    Return inside
        'End Function
    End Structure

    Function Plane_DotCoordinate(ByVal p As Plane, ByVal v As Vector3) As Single
        Return p.A * v.X + p.B * v.Y + p.C * v.Z + p.D
    End Function

    '  Transform an axis-aligned bounding box by the specified matrix, and compute a new axis-aligned bounding box
    Friend Function XFormBoundingBox(ByVal src As BoundingBox, ByVal mtx As Matrix) As BoundingBox
        '
        Dim pts(7) As Vector3
        For i As Int32 = 0 To 7
            pts(i) = src.Point_N(i)
        Next
        '
        Dim result As BoundingBox
        result.Init()
        '
        For i As Int32 = 0 To 7
            result.Merge(Vector3.TransformCoordinate(pts(i), mtx))
        Next
        '
        Return result
    End Function

    'Friend Sub BoundingBoxArray_PushBack(ByRef bbArray() As BoundingBox, ByVal bb As BoundingBox)
    '    ReDim Preserve bbArray(bbArray.Length)
    '    bbArray(bbArray.Length - 1) = New BoundingBox
    '    bbArray(bbArray.Length - 1).Init(bb)
    'End Sub



    ' ##############################################################################################
    ' ########
    ' ########        FRUSTUM
    ' ########
    ' ##############################################################################################
    ' ##############################################################################################
    ' ##############################################################################################

    Friend Structure Frustum
        Dim camPlanes() As Plane
        Dim nVertexLUT() As Int32
        Dim pntList() As Vector3

        ' -------------------------------  build a frustum from a camera (projection, or viewProjection) matrix
        Sub Init(ByVal m As Matrix)
            ReDim camPlanes(5)
            ReDim nVertexLUT(5)
            ReDim pntList(7)

            Dim col1 As Vector4 = VEC4(m.M11, m.M21, m.M31, m.M41)
            Dim col2 As Vector4 = VEC4(m.M12, m.M22, m.M32, m.M42)
            Dim col3 As Vector4 = VEC4(m.M13, m.M23, m.M33, m.M43)
            Dim col4 As Vector4 = VEC4(m.M14, m.M24, m.M34, m.M44)

            Dim planes(5) As Vector4
            planes(0) = col4 - col1  ' left
            planes(1) = col4 + col1  ' right
            planes(2) = col4 - col2  ' bottom
            planes(3) = col4 + col2  ' top
            planes(4) = col4 - col3  ' near
            planes(5) = col4 + col3  ' far
            ' ignore near & far plane

            Dim p As Int32
            ' ----------------------------- normalize the planes
            For p = 0 To 5
                Dim dot As Single = planes(p).X * planes(p).X + planes(p).Y * planes(p).Y + planes(p).Z * planes(p).Z
                dot = 1.0F / CSng(Math.Sqrt(dot))
                planes(p) = planes(p) * dot
            Next
            For p = 0 To 5
                camPlanes(p) = New Plane(planes(p).X, planes(p).Y, planes(p).Z, planes(p).W)
            Next

            '   build a bit-field that will tell us the indices for the nearest and farthest vertices from each plane...
            For i As Int32 = 0 To 5
                nVertexLUT(i) = 0
                If planes(i).X < 0 Then nVertexLUT(i) += 1
                If planes(i).Y < 0 Then nVertexLUT(i) += 2
                If planes(i).Z < 0 Then nVertexLUT(i) += 4
            Next

            ' ------------------------------------------------------------------------ compute extrema
            PlaneIntersection(pntList(0), camPlanes(5), camPlanes(2), camPlanes(1))
            PlaneIntersection(pntList(1), camPlanes(4), camPlanes(2), camPlanes(1))
            PlaneIntersection(pntList(2), camPlanes(5), camPlanes(3), camPlanes(1))
            PlaneIntersection(pntList(3), camPlanes(4), camPlanes(3), camPlanes(1))
            PlaneIntersection(pntList(4), camPlanes(5), camPlanes(2), camPlanes(0))
            PlaneIntersection(pntList(5), camPlanes(4), camPlanes(2), camPlanes(0))
            PlaneIntersection(pntList(6), camPlanes(5), camPlanes(3), camPlanes(0))
            PlaneIntersection(pntList(7), camPlanes(4), camPlanes(3), camPlanes(0))
        End Sub
        Function TestSphere(ByVal sphere As BoundingSphere) As Boolean
            For i As Int32 = 0 To 5
                If Plane_DotCoordinate(camPlanes(i), sphere.centerVec) + sphere.radius < 0 Then
                    Return False
                End If
            Next
            Return True
        End Function
        'Function TestSphere(ByVal sphere As BoundingSphere) As Boolean
        '    Dim inside As Boolean = True
        '    For i As Int32 = 0 To 5
        '        inside = inside And (Plane_DotCoordinate(camPlanes(i), sphere.centerVec) + sphere.radius >= 0)
        '        If Not inside Then Exit For
        '    Next
        '    Return inside
        'End Function
        ' -------------------------------------------------- Tests if an AABB is inside/intersecting the frustum
        Function TestBox(ByVal box As BoundingBox) As Int32
            Dim intersect As Boolean = False
            For i As Int32 = 0 To 5
                Dim nV As Int32 = nVertexLUT(i)
                ' --------------------------------------- pVertex is diagonally opposed to nVertex
                Dim nVertex As Vector3 = box.Point_N(nV)
                Dim pVertex As Vector3 = box.Point_P(nV)
                If Plane_DotCoordinate(camPlanes(i), nVertex) < 0.0F Then Return 0
                If Plane_DotCoordinate(camPlanes(i), pVertex) < 0.0F Then intersect = True
            Next
            If intersect Then
                Return 2
            Else
                Return 1
            End If
        End Function
        Function TestSweptSphere(ByVal sphere As BoundingSphere, ByVal sweepDir As Vector3) As Boolean

            '  get all 12 intersection points of the swept sphere with the view frustum
            '  for all points >0, displace sphere along the sweep driection.  
            '  if the displaced sphere is inside the frustum, return TRUE.  else, return FALSE

            Dim displacements(11) As Single
            Dim cnt As Int32 = 0
            Dim a, b As Single
            Dim inFrustum As Boolean = False

            For i As Int32 = 0 To 5
                If SweptSpherePlaneIntersect(a, b, camPlanes(i), sphere, sweepDir) Then
                    If (a >= 0.0F) Then
                        displacements(cnt) = a
                        cnt += 1
                    End If
                    If (b >= 0.0F) Then
                        displacements(cnt) = b
                        cnt += 1
                    End If
                End If
            Next

            For i As Int32 = 0 To cnt - 1
                Dim displacedSphere As BoundingSphere
                displacedSphere.Init(sphere)
                displacedSphere.centerVec += sweepDir * displacements(i)
                displacedSphere.radius *= 1.1F
                If TestSphere(displacedSphere) Then inFrustum = True
            Next
            Return inFrustum
        End Function
    End Structure

    Function SweptSpherePlaneIntersect(ByRef t0 As Single, ByRef t1 As Single, ByVal _plane As Plane, ByVal sphere As BoundingSphere, ByVal sweepDir As Vector3) As Boolean
        Dim b_dot_n As Single = Plane_DotCoordinate(_plane, sphere.centerVec)
        Dim d_dot_n As Single = Plane.DotNormal(_plane, sweepDir)
        If d_dot_n = 0.0F Then
            If b_dot_n <= sphere.radius Then
                t0 = 0.0F
                t1 = 3.3E+33F ' effectively infinity
                Return True
            Else
                Return False
            End If
        Else
            Dim tmp0 As Single = (sphere.radius - b_dot_n) / d_dot_n
            Dim tmp1 As Single = (-sphere.radius - b_dot_n) / d_dot_n
            t0 = Math.Min(tmp0, tmp1)
            t1 = Math.Max(tmp0, tmp1)
            Return True
        End If
    End Function

    ' computes the point where three planes intersect and returns whether or not the point exists
    Function PlaneIntersection(ByRef intersectPt As Vector3, ByVal p0 As Plane, ByVal p1 As Plane, ByVal p2 As Plane) As Boolean
        Dim n0 As Vector3 = New Vector3(p0.A, p0.B, p0.C)
        Dim n1 As Vector3 = New Vector3(p1.A, p1.B, p1.C)
        Dim n2 As Vector3 = New Vector3(p2.A, p2.B, p2.C)

        Dim n1_n2 As Vector3 = Vector3.Cross(n1, n2)
        Dim n2_n0 As Vector3 = Vector3.Cross(n2, n0)
        Dim n0_n1 As Vector3 = Vector3.Cross(n0, n1)

        Dim cosTheta As Single = Vector3.Dot(n0, n1_n2)

        If ALMOST_ZERO(cosTheta) Or IS_SPECIAL(cosTheta) Then Return False

        Dim secTheta As Single = 1.0F / cosTheta

        n1_n2 = n1_n2 * p0.D
        n2_n0 = n2_n0 * p1.D
        n0_n1 = n0_n1 * p2.D

        intersectPt = -(n1_n2 + n2_n0 + n0_n1) * secTheta
        Return True
    End Function


    ' ##############################################################################################
    ' ########
    ' ########        BOUNDING SPHERE
    ' ########
    ' ##############################################################################################
    ' ##############################################################################################
    ' ##############################################################################################

    Friend Structure BoundingSphere
        Dim centerVec As Vector3
        Dim radius As Single

        Sub Init(ByVal _centerVec As Vector3, ByVal _radius As Single)
            centerVec = _centerVec
            radius = _radius
        End Sub
        Sub Init(ByVal bs As BoundingSphere)
            centerVec = bs.centerVec
            radius = bs.radius
        End Sub
        Sub Init(ByVal points() As Vector3)
            radius = 0
            centerVec = points(0)
            Dim d As Single
            Dim cVec As Vector3
            Dim i As Int32 = 1
            While i < points.Length
                cVec = points(i) - centerVec
                d = Vector3.Dot(cVec, cVec)
                If d > radius * radius Then
                    d = CSng(Math.Sqrt(d))
                    Dim r As Single = 0.5F * (d + radius)
                    Dim scale As Single = (r - radius) / d
                    centerVec = centerVec + scale * cVec
                    radius = r
                End If
                i += 1
            End While
        End Sub
        Sub Init(ByVal box As BoundingBox)
            centerVec = 0.5F * (box.maxPt + box.minPt)
            radius = (box.maxPt - centerVec).Length
        End Sub
    End Structure



    '#define FLT_AS_DW(F) (*(DWORD*)&(F))
    '#define ALMOST_ZERO(F) ((FLT_AS_DW(F) & 0x7f800000L)==0)
    '#define IS_SPECIAL(F)  ((FLT_AS_DW(F) & 0x7f800000L)==0x7f800000L)

    Friend Function ALMOST_ZERO(ByVal n As Single) As Boolean
        ' ---------------------------------------------------------------- 1
        'Return (CType(n, Int32) And &H7F800000) = 0

        ' ---------------------------------------------------------------- 2
        'If n < 1.0E-33 Then
        '    Return True
        'Else
        '    Return False
        'End If

        ' ---------------------------------------------------------------- 3
        If Math.Abs(n) < 1.0E-30 Then
            Return True
        Else
            Return False
        End If

    End Function

    Friend Function IS_SPECIAL(ByVal n As Single) As Boolean
        ' ---------------------------------------------------------------- 1
        'Return (CType(n, Int32) And &H7F800000) = &H7F800000
        Return False
    End Function



    ' ==============================================================================================================
    '   COMBO FUNCTIONS
    ' ==============================================================================================================

    Friend Sub Combo_Init(ByVal combo As ComboBox, ByVal str As String)
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        If str = Nothing Then str = ""
        With combo
            .Items.Clear()
            .Items.Add(str)
            .SelectedIndex = 0
        End With
        EventsAreEnabled = old
    End Sub

    Friend Sub Combo_SetIndex_FromString(ByVal combo As ComboBox, ByVal str As String)
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        If str = Nothing Then str = ""
        With combo
            For i As Int32 = 0 To .Items.Count - 1
                If .Items(i).ToString = str Then
                    .SelectedIndex = i
                    Exit For
                End If
            Next
        End With
        EventsAreEnabled = old
    End Sub

    Friend Function Combo_GetValue(ByVal combo As ComboBox) As String
        If combo.SelectedIndex < 0 Then Return ""
        Return combo.Items(combo.SelectedIndex).ToString()
    End Function

    Friend Sub Combo_SetIndex(ByVal combo As ComboBox, ByVal index As Int32)
        If combo.Items.Count < 1 Then Exit Sub
        If index < 0 Then index = 0
        If index > combo.Items.Count - 1 Then index = combo.Items.Count - 1
        combo.SelectedIndex = index
    End Sub

End Module
