﻿


Module Module_SaveLoad

    Friend EventsAreEnabled As Boolean = False

    Friend Form1_WindowStateAtStart As String
    Friend Form2_VisibleAtStart As Boolean
    Friend Application_Closing As Boolean

    ' =======================================================================================================
    '   APP TITLE AND VERSION
    ' =======================================================================================================
    Friend Function AppTitleAndVersion(Optional ByVal Title As String = "") As String
        If Title = "" Then Title = Replace(My.Application.Info.AssemblyName, "_", " ")
        Dim s() As String = Split(My.Application.Info.Version.ToString, ".")
        Return Title & " - V" & s(0) & "." & s(1)
    End Function


    Private Function GetValueInt(ByVal s As String) As Int32
        Return CInt(Val(s))
    End Function

    Private Function GetValue(ByVal s As String) As Single
        Return CSng(Val(Replace(s, ",", ".")))
    End Function

    ' =======================================================================================
    '  LimitFormPosition
    ' =======================================================================================
    Friend Sub LimitFormToPrimaryScreen_CompletelyVisible(ByVal f As System.Windows.Forms.Form)
        SB.X = 0
        SB.Y = 0
        SB.Width = Screen.PrimaryScreen.Bounds.Width
        SB.Height = Screen.PrimaryScreen.Bounds.Height
        EnsureFormCompletelyVisible(f)
    End Sub

    Friend Sub LimitFormToAllScreens(ByVal f As System.Windows.Forms.Form)
        If f.WindowState <> FormWindowState.Normal Then Return
        GetMaxScreenBounds()
        EnsureFormVisible(f)
    End Sub

    Private SB As Rectangle = New Rectangle(Integer.MaxValue, Integer.MaxValue, Integer.MinValue, Integer.MinValue)

    Private Sub GetMaxScreenBounds()
        For Each s As Screen In System.Windows.Forms.Screen.AllScreens
            SB = Rectangle.Union(SB, s.WorkingArea)
        Next
    End Sub

    Private Sub EnsureFormCompletelyVisible(ByVal frm As Form)
        With frm
            .Width = Math.Min(.Width, SB.Width)         ' not more than a maximized window
            .Height = Math.Min(.Height, SB.Height)      ' not more than a maximized window
            .Width = Math.Max(.Width, 32)               ' at least 32x24
            .Height = Math.Max(.Height, 24)             ' at least 32x24
            .Left = Math.Min(.Left, SB.Right - .Width)  ' not beyond the right border
            .Top = Math.Min(.Top, SB.Bottom - .Height)  ' not beyond the bottom border
            .Left = Math.Max(.Left, SB.Left)            ' at least at the left border
            .Top = Math.Max(.Top, SB.Top)               ' at least at the top border
        End With
    End Sub

    Private Sub EnsureFormVisible(ByVal frm As Form)
        With frm
            .Width = Math.Min(.Width, SB.Width)             ' not more than VIRTUALSCREEN dimensions
            .Height = Math.Min(.Height, SB.Height)          ' not more than VIRTUALSCREEN dimensions 
            .Width = Math.Max(.Width, 32)                   ' at least 32x24
            .Height = Math.Max(.Height, 24)                 ' at least 32x24
            .Left = Math.Min(.Left, SB.Right - 50)          ' not beyond right border - 50 pixels
            .Top = Math.Min(.Top, SB.Bottom - 100)          ' not beyond bottom border - 50 pixels
            .Left = Math.Max(.Left, SB.Left + 100 - .Width) ' at least at left border + 50 pixels
            .Top = Math.Max(.Top, SB.Top - 10)              ' at least at top border
        End With
    End Sub

    ' (The value of the RestoreBounds property is valid only 
    '   when the WindowState property of the Form class is not equal to Normal)
    Friend Function GetFormRectangle(ByVal frm As Form) As Rectangle
        Dim r As Rectangle
        If frm.WindowState = FormWindowState.Normal Then
            r = frm.Bounds
        Else
            r = frm.RestoreBounds
        End If
        Return r
    End Function


    ' ================================================================================================
    '  Private Read-Write functions
    ' ================================================================================================
    Private Function TabString(ByVal Name As String, _
                              Optional ByVal Value As Double = Double.NaN, _
                              Optional ByVal fmt As String = "") As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)
        If Double.IsNaN(Value) Then
            Return Name
        Else
            Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString(fmt)
        End If
    End Function

    Private Function TabString(ByVal Name As String, _
                                  ByVal Value As Boolean) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString
    End Function

    Private Function TabString(ByVal Name As String, _
                                ByVal Value As String) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value
    End Function

    Private Function Val_Double(ByVal l As String) As Double
        Return Val(l.Replace(",", "."))
    End Function

    Private Function Val_Int(ByVal l As String) As Int32
        Return CInt(Val(l))
    End Function

    Private Function ExtractParamName(ByRef s As String) As String
        ' ------------------------- Returns the first field from begin to the first "=" symbol
        ' -------------------------  and removes it from the string
        Dim i As Integer
        i = InStr(s, "=")
        If i > 0 Then
            ExtractParamName = Trim(Strings.Left(s, i - 1))
            s = Trim(Mid(s, i + 1))
        Else
            ExtractParamName = Trim(s)
            s = ""
        End If
    End Function

    Private Function AssemblyName() As String
        Return System.Reflection.Assembly.GetExecutingAssembly.GetName.Name
    End Function

    ' ==================================================================================================
    '  SAVE LOAD -- Program INI
    ' ==================================================================================================

    Friend Sub Save_INI()
        '
        Dim iniFileName As String = Application.StartupPath & "\" & AssemblyName() & "_INI.txt"
        Dim f As System.IO.StreamWriter = Nothing
        Try
            f = IO.File.CreateText(iniFileName)
            '
            f.WriteLine(" Program Params")
            f.WriteLine("===========================================")
            Dim r As Rectangle
            r = GetFormRectangle(Form1)
            f.WriteLine(TabString("Form1_Top", r.Top))
            f.WriteLine(TabString("Form1_Left", r.Left))
            f.WriteLine(TabString("Form1_Width", r.Width))
            f.WriteLine(TabString("Form1_Height", r.Height))
            f.WriteLine(TabString("Form1_WindowState", Form1.WindowState.ToString))
            '
            r = GetFormRectangle(Form2)
            f.WriteLine(TabString("Form2_Top", r.Top))
            f.WriteLine(TabString("Form2_Left", r.Left))
            f.WriteLine(TabString("Form2_Visible", Form2.Visible.ToString))
            '
            f.WriteLine("")
            f.WriteLine(TabString(" App"))
            f.WriteLine(TabString("==========================================="))
            f.WriteLine(TabString("App_CurrentScene", App_CurrentScene.ToString))
            f.WriteLine(TabString("App_LightDirection", DX_LightDirection.ToString))
            f.WriteLine(TabString("App_LightHeight", DX_LightHeight.ToString))
            f.WriteLine(TabString("App_SkyType", DX_SkyType.ToString))
            '
            f.WriteLine(TabString(""))
            f.WriteLine(TabString(" Shadows"))
            f.WriteLine(TabString("==========================================="))
            f.WriteLine(TabString("Shadow_Bias", SHAD_Bias.ToString("0.000")))
            f.WriteLine(TabString("Shadow_BiasAutoRange", SHAD_BiasAutoRange.ToString("0.000")))
            f.WriteLine(TabString("Shadow_Coef", SHAD_Coefficent.ToString("0.000")))
            f.WriteLine(TabString("Shadow_EpsilonW", SHAD_EpsilonW.ToString("0.000")))
            f.WriteLine(TabString("Shadow_MaxDistance", SHAD_MaxDistance.ToString))
            f.WriteLine(TabString("Shadow_FilterGain", SHAD_FilterGain.ToString("0.000")))
            f.WriteLine(TabString("Shadow_FilterType", SHAD_FilterType.ToString))
            f.WriteLine(TabString("Shadow_Density", SHAD_ShadowDensity.ToString("0.000")))
            f.WriteLine(TabString("Shadow_ShadowMapSize", SHAD_ShadowMapSize.ToString))
            f.WriteLine(TabString("Shadow_FinalMapSize", SHAD_FinalMapSize.ToString))
            f.WriteLine(TabString("Shadow_XpsmCorrectedRotation", SHAD_XpsmCorrectedRotation.ToString))
            f.WriteLine(TabString("Shadow_ShadowAddSmooth", SHAD_ShadowAddSmooth.ToString))
            f.WriteLine(TabString("Shadow_HiQualityFilters", SHAD_HiQualityFilters.ToString))
            f.WriteLine(TabString("Shadow_HiQualityObjects", SHAD_HiQualityObjects.ToString))
            '
            f.WriteLine(TabString(""))
            f.WriteLine(TabString(" Device"))
            f.WriteLine(TabString("==========================================="))
            f.WriteLine(TabString("Device_BackBufferWidth", DEVICE_BackBufferWidth.ToString))
            f.WriteLine(TabString("Device_BackBufferHeight", DEVICE_BackBufferHeight.ToString))
            f.WriteLine(TabString("Device_Multisamples", DEVICE_Multisamples.ToString))
            f.WriteLine(TabString("Device_HardwareVertex", DEVICE_HardwareVertexProcessing.ToString))
            '
            f.WriteLine(TabString(""))
            f.WriteLine(TabString(" Debug"))
            f.WriteLine(TabString("==========================================="))
            f.WriteLine(TabString("Debug_ViewMainParams", DEBUG_ViewMainParams.ToString))
            f.WriteLine(TabString("Debug_ViewShadowMap", DEBUG_ViewShadowMap.ToString))
            f.WriteLine(TabString("Debug_ViewFinalMap", DEBUG_ViewFinalMap.ToString))
            f.WriteLine(TabString("Debug_DisableShadows", DEBUG_DisableShadows.ToString))
            f.WriteLine(TabString("Debug_DisableSounds", DEBUG_DisableSounds.ToString))
            f.WriteLine(TabString("Debug_HiQuality", DEBUG_HiQuality.ToString))
            f.WriteLine(TabString("Debug_DrawWireFrame", DEBUG_DrawWireFrame.ToString))
            f.WriteLine(TabString("Debug_RunInWindow", DEBUG_RunInWindow.ToString))
            f.WriteLine(TabString("Debug_UseThreads", DEBUG_UseThreads.ToString))
            f.WriteLine(TabString("Debug_ShowMovingObjects", DEBUG_ShowMovements.ToString))
            f.WriteLine(TabString("Debug_ViewDebugParams", DEBUG_ViewDebugParams.ToString))
            '
            f.WriteLine("")
            f.WriteLine(" InOut")
            f.WriteLine("===========================================")
            f.WriteLine(TabString("SlotDir", Form2.txt_SlotDir.Text))
            f.WriteLine(TabString("SlotUp", Form2.txt_SlotUp.Text))
            f.WriteLine(TabString("SlotLbutton", Form2.txt_SlotLbutton.Text))
            f.WriteLine(TabString("SlotRbutton", Form2.Label26.Text))
            '
        Catch
        End Try
        Try
            f.Close()
        Catch
        End Try
    End Sub

    Friend Sub Load_INI()
        ' ------------------------------------------------------------- default values

        ' ------------------------------------------------------------- 
        ' ------------------------------------------------------------------------------- 
        ' With "Resume Next" subsequent parameters are loaded and f.Close() is executed
        ' -------------------------------------------------------------------------------
        On Error Resume Next  ' use Resume-Next instead of Try-Catch
        ' -------------------------------------------------------------------------------
        Dim l As String
        Dim iniFileName As String = Application.StartupPath & "\" & AssemblyName() & "_INI.txt"
        If My.Computer.FileSystem.FileExists(iniFileName) Then
            Dim f As System.IO.StreamReader
            f = IO.File.OpenText(iniFileName)
            Do While Not f.EndOfStream
                l = f.ReadLine()
                Select Case ExtractParamName(l)
                    ' ------------------------------------------------------------------- Form1
                    Case "Form1_Top" : Form1.Top = GetValueInt(l)
                    Case "Form1_Left" : Form1.Left = GetValueInt(l)
                    Case "Form1_Width" : Form1.Width = GetValueInt(l)
                    Case "Form1_Height" : Form1.Height = GetValueInt(l)
                    Case "Form1_WindowState" : Form1_WindowStateAtStart = l
                        ' ------------------------------------------------------------------- Form2
                    Case "Form2_Top" : Form2.Top = GetValueInt(l)
                    Case "Form2_Left" : Form2.Left = GetValueInt(l)
                    Case "Form2_Visible" : Form2_VisibleAtStart = l = "True"
                        ' --------------------------------------------------------------- Main
                    Case "App_CurrentScene" : App_CurrentScene = GetValueInt(l)
                    Case "App_LightDirection" : DX_LightDirection = GetValueInt(l)
                    Case "App_LightHeight" : DX_LightHeight = GetValueInt(l)
                    Case "App_SkyType" : DX_SkyType = GetValueInt(l)
                        ' --------------------------------------------------------------- Shadows
                    Case "Shadow_Bias" : SHAD_Bias = GetValue(l)
                    Case "Shadow_BiasAutoRange" : SHAD_BiasAutoRange = GetValue(l)
                    Case "Shadow_Coef" : SHAD_Coefficent = GetValue(l)
                    Case "Shadow_EpsilonW" : SHAD_EpsilonW = GetValue(l)
                    Case "Shadow_MaxDistance" : SHAD_MaxDistance = GetValue(l)
                    Case "Shadow_FilterGain" : SHAD_FilterGain = GetValue(l)
                    Case "Shadow_FilterType" : SHAD_FilterType = GetValueInt(l)
                    Case "Shadow_Density" : SHAD_ShadowDensity = GetValue(l)
                    Case "Shadow_ShadowMapSize" : SHAD_ShadowMapSize = GetValueInt(l)
                    Case "Shadow_FinalMapSize" : SHAD_FinalMapSize = GetValueInt(l)
                    Case "Shadow_XpsmCorrectedRotation" : SHAD_XpsmCorrectedRotation = l = "True"
                    Case "Shadow_ShadowAddSmooth" : SHAD_ShadowAddSmooth = l = "True"
                    Case "Shadow_HiQualityFilters" : SHAD_HiQualityFilters = l = "True"
                    Case "Shadow_HiQualityObjects" : SHAD_HiQualityObjects = l = "True"
                        ' --------------------------------------------------------------- Device
                    Case "Device_BackBufferWidth" : DEVICE_BackBufferWidth = GetValueInt(l)
                    Case "Device_BackBufferHeight" : DEVICE_BackBufferHeight = GetValueInt(l)
                    Case "Device_Multisamples" : DEVICE_Multisamples = GetValueInt(l)
                    Case "Device_HardwareVertex" : DEVICE_HardwareVertexProcessing = l = "True"
                        ' --------------------------------------------------------------- Debug
                    Case "Debug_ViewMainParams" : DEBUG_ViewMainParams = l = "True"
                    Case "Debug_ViewShadowMap" : DEBUG_ViewShadowMap = l = "True"
                    Case "Debug_ViewFinalMap" : DEBUG_ViewFinalMap = l = "True"
                    Case "Debug_DisableShadows" : DEBUG_DisableShadows = l = "True"
                    Case "Debug_DisableSounds" : DEBUG_DisableSounds = l = "True"
                    Case "Debug_HiQuality" : DEBUG_HiQuality = l = "True"
                    Case "Debug_DrawWireFrame" : DEBUG_DrawWireFrame = l = "True"
                    Case "Debug_RunInWindow" : DEBUG_RunInWindow = l = "True"
                    Case "Debug_UseThreads" : DEBUG_UseThreads = l = "True"
                    Case "Debug_ShowMovingObjects" : DEBUG_ShowMovements = l = "True"
                    Case "Debug_ViewDebugParams" : DEBUG_ViewDebugParams = l = "True"
                        '

                    Case "SlotDir" : Form2.txt_SlotDir.Text = l
                    Case "SlotUp" : Form2.txt_SlotUp.Text = l
                    Case "SlotLbutton" : Form2.txt_SlotLbutton.Text = l
                    Case "SlotRbutton" : Form2.Label26.Text = l
                End Select
            Loop
            f.Close()
        End If
        LimitFormToAllScreens(Form2)
    End Sub

End Module
