#pragma once

#include "btBulletDynamicsCommon.h"

#define COLLISION_RADIUS 0.0f

// Conversion helpers
D3DXMATRIX		ConvertMatrix( btTransform &trn );
D3DXVECTOR3		ConvertVector( btVector3 &vec );
btVector3		ConvertVector( const D3DXVECTOR3 *vec );
btVector3		ConvertVector( D3DXVECTOR3 &vec );

// Must store the collision shape returned by these functions so we can delete later on
btCollisionShape * CreatePyramid(btDynamicsWorld *world, const btVector3& offsetPosition, int stackSize, const btVector3& shapeSize, int shapeType);
btCollisionShape * CreateWall(btDynamicsWorld *world, const btVector3& offsetPosition, int stackSize,const btVector3& boxSize);
btCollisionShape * CreateTowerCircle(btDynamicsWorld *pWorld, const btVector3& offsetPosition, int stackSize, int rotSize, const btVector3& boxSize);
btRigidBody		*CreateRigidBody(float mass, const btTransform& startTransform, btCollisionShape* shape);

struct TriMeshData
{
	btScalar	*vertices;
	int			*indices;	
	btTriangleIndexVertexArray	*indexVertexArrays;
	TriMeshData()
	{
		vertices = NULL;
		indices = NULL;
		indexVertexArrays = NULL;
	}
	~TriMeshData()
	{
		if( vertices ) delete [] vertices;
		if( indices ) delete [] indices;
		if( indexVertexArrays ) delete indexVertexArrays;
	}
};

btBvhTriangleMeshShape *CreateBvhTriangleMeshShape(LPD3DXMESH pMesh, TriMeshData *pData);
btConvexHullShape *CreateConvexHullShape( LPD3DXMESH pMesh );
