#pragma once

#include "Utilities.h"

class HeightMap
{
public:
	HeightMap( Point size, float maxheight );
	~HeightMap(void);

	void Release();

	// Blends 2 height map together
	void operator *= (const HeightMap &rhs);

	// Random generate a height map
	HRESULT CreateRandomHeightMap( int seed, float noiseSize, float persistence, int octaves);
	// Loads the height map from a texture file
	HRESULT LoadFromFile(LPDIRECT3DDEVICE9 device, char fileName[]);
	
	// Puts the height map into a texture, useful for debugging
	HRESULT GenerateHeightMapTexture(LPDIRECT3DDEVICE9 device, LPDIRECT3DTEXTURE9 *texture);  
	
	// Raise / lower a rectangular shape in the heightmap
	void Raise(RECT r, float f);
	
	// Raise / lower a circle shape
	void Raise( Point pt, int radius, float f );
	
	// Smooths the height map
	void Smooth();
	
	// Clamps the height to a certain height limit
	void Cap(float capHeight);	

	float *GetHeightTable()						{ return mHeightTable; }
	float GetMinHeight()						{ return mNewMinHeight; }
	float GetMaxHeight()						{ return mNewMaxHeight; }
	Point GetSize()								{ return mSize; }
	float GetHeightAt( int x, int y )			{ return mHeightTable[ x + y * mSize.x ]; }
	void ModifyHeightAt( int x, int y, float f ){ mHeightTable[ x + y * mSize.x ] += f; }
	void SetHeightAt( int x, int y, float f )	{ mHeightTable[ x + y * mSize.x ] = f; }
	bool Within(Point p);

private:	
	Point mSize;	
	float mMaxHeight;	
	float mNewMinHeight;
	float mNewMaxHeight;
	float *mHeightTable;		
};
