#pragma once

class btCollisionShape;
class btConcaveShape;

#include "LinearMath/btAlignedObjectArray.h"
#include "LinearMath/btVector3.h"

#include "BulletCollision/CollisionShapes/btShapeHull.h"
#include "BulletCollision/CollisionShapes/btTriangleCallback.h"

#include "Vertex.h"
#include <vector>

class btSoftBody;

class ShapeDrawer
{
	struct ShapeCache
	{
		struct Edge { btVector3 n[2]; int v[2]; };
		ShapeCache(btConvexShape* s) : shapeHull(s) {}
		btShapeHull					shapeHull;
		btAlignedObjectArray<Edge>	edges;
	};	

	struct ConcaveShapeCache
	{
		LPDIRECT3DVERTEXBUFFER9 vertexBuffer;
		int numFaces;
	};

public:
	ShapeDrawer(void);
	~ShapeDrawer(void);

	void Create( LPDIRECT3DDEVICE9 device );
	void Reset( LPDIRECT3DDEVICE9 device );
	void Lost();
	void Destroy();
	void AddConcaveShape( btConcaveShape *shape);

	void Render( const btCollisionShape* shape, btTransform &trn);		
	void RenderConcaveShapes( D3DXMATRIX *world, size_t index );	

	void ResetStats()		{ mNumVertices = mNumFaces = mNumObjects = 0; }
	DWORD GetNumVertices()	{ return mNumVertices; }
	DWORD GetNumFaces()		{ return mNumFaces; }
	DWORD GetNumObjects()	{ return mNumObjects; }

private:
	ShapeCache*	Cache(btConvexShape*);

private:
	LPDIRECT3DDEVICE9	mDevice;
	
	// Use for drawing Bullet primitive shapes 
	LPD3DXMESH			mBoxMesh;
	LPD3DXMESH			mSphereMesh;
	LPD3DXMESH			mCylinderMesh;
	LPD3DXMESH			mConeMesh;
	
	btAlignedObjectArray<ShapeCache*>	mShapeCaches;
	std::vector<ConcaveShapeCache *>	mConcaveShapeCaches;

	DWORD				mNumVertices;
	DWORD				mNumFaces;
	DWORD				mNumObjects;	
};

class DxDrawcallback : public btTriangleCallback
{
public:
	//VertexNormal *vertices;
	std::vector<VertexNormal> vertices;
	
	int verts;
	int tri ;

	DxDrawcallback()
	{
      verts = 0;
	  tri = 0;	 	 
	}

	~DxDrawcallback()
	{
		vertices.clear();
	}

	virtual void processTriangle(btVector3* triangle,int partId, int triangleIndex)
	{	
		VertexNormal tempvertices[3] =
		{
			{ D3DXVECTOR3( triangle[0].getX(), triangle[0].getY(), triangle[0].getZ() ), D3DXVECTOR3(0.0f, 0.0f, 0.0f) },
			{ D3DXVECTOR3( triangle[1].getX(), triangle[1].getY(), triangle[1].getZ() ), D3DXVECTOR3(0.0f, 0.0f, 0.0f) },
			{ D3DXVECTOR3( triangle[2].getX(), triangle[2].getY(), triangle[2].getZ() ), D3DXVECTOR3(0.0f, 0.0f, 0.0f) },
		};		
		
		// Generate normal
		D3DXVECTOR3 normal;
		D3DXVECTOR3 vec1 = tempvertices[2].position - tempvertices[0].position;
		D3DXVECTOR3 vec2 = tempvertices[0].position - tempvertices[1].position;
		D3DXVec3Cross( &normal, &vec1, &vec2 );
		D3DXVec3Normalize( &normal, &normal );

		tempvertices[0].normal = normal;
		tempvertices[1].normal = normal;
		tempvertices[2].normal = normal;
		
		vertices.push_back( tempvertices[0] );
		vertices.push_back( tempvertices[1] );
		vertices.push_back( tempvertices[2] );
	
		verts +=3;
		tri++;
	}
};

