
// BulletWrapper.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"
#include "BulletWrapper.h"
//#include "DXUT.h"
//#include "Utils.h"

#include <btBulletDynamicsCommon.h>
#include <BulletSoftBody/btSoftBody.h>
//#include <BulletSoftBody/btSoftBodyInternals.h>
#include <BulletSoftBody/btSoftRigidDynamicsWorld.h>
#include "BulletSoftBody/btSoftBodyRigidBodyCollisionConfiguration.h"
#include <BulletSoftBody/btSoftBodyHelpers.h>

//#include "../Bullet-2.77\src\BulletCollision\CollisionDispatch\btSimulationIslandManager.h"

//#include <LinearMath/btQuickprof.h>
//#include <BulletCollision/CollisionDispatch/btGhostObject.h>
//#include <BulletCollision/CollisionShapes/btHeightfieldTerrainShape.h>
//#include <BulletDynamics/Character/btKinematicCharacterController.h>



//#define USE_PARALLEL_DISPATCHER
#ifdef USE_PARALLEL_DISPATCHER
#include "../Bullet-2.77\src/BulletMultiThreaded/SpuGatheringCollisionDispatcher.h"
#include "../Bullet-2.77\src/BulletMultiThreaded/Win32ThreadSupport.h"
#include "../Bullet-2.77\src/BulletMultiThreaded/SpuLibspe2Support.h"
#include "../Bullet-2.77\src/BulletMultiThreaded/SpuNarrowPhaseCollisionTask/SpuGatheringCollisionTask.h"
#endif



// This is the constructor of a class that has been exported.
// see BulletWrapper.h for the class definition
//CBulletWrapper::CBulletWrapper()
//{
//	return;
//}


#define SAFE_EXIT(i) { if ( (i) < 0 || (i) >= mWorld->getNumCollisionObjects() ) return; }
#define SAFE_EXIT_INT(i) { if ( (i) < 0 || (i) >= mWorld->getNumCollisionObjects() ) return 0; }
#define SAFE_EXIT_FLOAT(i) { if ( (i) < 0 || (i) >= mWorld->getNumCollisionObjects() ) return 0.0f; }
#define SAFE_EXIT_MATRIX(i) { if ( (i) < 0 || (i) >= mWorld->getNumCollisionObjects() ) return D3DXMATRIX(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1); }
#define SAFE_EXIT_VEC3(i) { if ( (i) < 0 || (i) >= mWorld->getNumCollisionObjects() ) return D3DXVECTOR3(0,0,0); }

// ===========================================================================================
//   GLOBALS
// ===========================================================================================

btSoftRigidDynamicsWorld*				mWorld;
btDefaultCollisionConfiguration*		mCollisionConfiguration;
btSequentialImpulseConstraintSolver*	mConstraintSolver;
btCollisionDispatcher*					mCollisionDispatcher;	
btBroadphaseInterface*					mBroadPhase;
btSoftBodyWorldInfo						mSoftBodyWorldInfo;

btAlignedObjectArray<btCollisionShape*> mCollisionShapes;	





// ===========================================================================================
//   EXPORTED FUNCTIONS
//
//  The normal CallingConvention ( Declare in VB ) is Winapi, which defaults to StdCall
// ===========================================================================================

// =======================================================================================
//   INIT / DESTROY / STEP
// =======================================================================================

BULLETWRAPPER_API void Bullet_Init(D3DXVECTOR3 WorldMin, 
								   D3DXVECTOR3 WorldMax, 
								   D3DXVECTOR3 Gravity)
{

	//register some softbody collision algorithms on top of the default btDefaultCollisionConfiguration
	mCollisionConfiguration = new btSoftBodyRigidBodyCollisionConfiguration();


#ifdef USE_PARALLEL_DISPATCHER
	// see: http://bullet.googlecode.com/svn/tags/bullet-2.72-release/Demos/CellSpuDemo/BasicDemo2.cpp

	__int32 maxNumOutstandingTasks = 1;
	Win32ThreadSupport* threadSupport = new Win32ThreadSupport(Win32ThreadSupport::Win32ThreadConstructionInfo(
																						"collision",
																						processCollisionTask,
																						createCollisionLocalStoreMemory,
																						maxNumOutstandingTasks));
	mCollisionDispatcher = new SpuGatheringCollisionDispatcher(threadSupport,
															   maxNumOutstandingTasks,
															   mCollisionConfiguration);
	mCollisionConfiguration = new btSoftBodyRigidBodyCollisionConfiguration();
#else
	mCollisionDispatcher = new btCollisionDispatcher(mCollisionConfiguration);
#endif

	// ---------------------------------------------------------------------------------- BroadPhase
	//mBroadPhase = new btAxisSweep3(ConvertVector(WorldMin), ConvertVector(WorldMax));
	mBroadPhase = new btDbvtBroadphase();
	//mBroadPhase = new btMultiSapBroadphase();
	//mBroadPhase = new btSimpleBroadphase();

	// ---------------------------------------------------------------------------------- ConstraintSolver
	mConstraintSolver = new btSequentialImpulseConstraintSolver();

	// ---------------------------------------------------------------------------------- World
	mWorld = new btSoftRigidDynamicsWorld(mCollisionDispatcher, 
										  mBroadPhase, 
										  mConstraintSolver, 
										  mCollisionConfiguration);

	// ---------------------------------------------------------------------------------------------------
	//http://bulletphysics.org/mediawiki-1.5.8/index.php/BtContactSolverInfo
	// ---------------------------------------------------------------------------------------------------
	//mWorld->getSolverInfo().m_numIterations = 4;  // default = 10  ( 4 - low quality to 20 - good quality )
	//mWorld->getSolverInfo().m_solverMode = SOLVER_SIMD + SOLVER_USE_WARMSTARTING;  // <<<< default
	//mWorld->getSolverInfo().m_solverMode = SOLVER_SIMD+SOLVER_RANDMIZE_ORDER+SOLVER_USE_WARMSTARTING;
	//mWorld->getSolverInfo().m_solverMode = SOLVER_CACHE_FRIENDLY | SOLVER_RANDMIZE_ORDER;
	//mWorld->getSolverInfo().m_solverMode = SOLVER_SIMD + SOLVER_CUDA + SOLVER_USE_WARMSTARTING;
	//mWorld->getSolverInfo().m_solverMode = SOLVER_USE_WARMSTARTING;

	// ---------------------------------------------------------------------------------------------------
	//By default, Bullet solves positional constraints and velocity constraints coupled together.
    //This works well in many cases, but the error reduction of position coupled to velocity
	//introduces extra energy (noticeable as 'bounce'). 
	//Instead of coupled positional and velocity constraint solving, the two can be solved separately
	//using the 'split impulse' option. 
	//This means that recovering from deep penetrations doesn't add any velocity. 
	//You can enable the option using: 
	// mWorld->getSolverInfo().m_splitImpulse = 0; //enable split impulse feature
	// mWorld->getSolverInfo().m_splitImpulsePenetrationThreshold = -0.02;
	// ---------------------------------------------------------------------------------------------------
	mWorld->getSolverInfo().m_splitImpulse = 0; // ensure split impulse disabled


	//mWorld->getSimulationIslandManager()->setSplitIslands(true);  // default = true

	      
	mWorld->getDispatchInfo().m_enableSPU = true;


	mWorld->setGravity(ConvertVector(Gravity));
	mSoftBodyWorldInfo.m_gravity.setValue(Gravity.x, Gravity.y, Gravity.z);

	mSoftBodyWorldInfo.m_sparsesdf.Initialize();

	mSoftBodyWorldInfo.air_density		= 1.2F;
	mSoftBodyWorldInfo.water_density	= 0;
	mSoftBodyWorldInfo.water_offset		= 0;
	mSoftBodyWorldInfo.water_normal		= btVector3(0,0,0);

	mSoftBodyWorldInfo.m_dispatcher = mCollisionDispatcher;
	mSoftBodyWorldInfo.m_broadphase = mBroadPhase;


}






// ==================================================================== 2 - ( FROM BULLET DEMOS )
BULLETWRAPPER_API void Bullet_Destroy()
{
	if( mWorld )
	{
		for(int i = mWorld->getNumCollisionObjects()-1; i >= 0; i--)
		{
			btCollisionObject*	obj=mWorld->getCollisionObjectArray()[i];
			btRigidBody*		body=btRigidBody::upcast(obj);
			if(body)
			{
				btActionInterface* vehicle = (btActionInterface*)body->m_vehicle;
				if (vehicle) 
				{
					mWorld->removeVehicle(vehicle);
					SAFE_DELETE(vehicle);
				}
				if (body->getMotionState())
				{
					delete body->getMotionState();
				}
			}
			while(mWorld->getNumConstraints())
			{
				btTypedConstraint* pc = mWorld->getConstraint(0);
				mWorld->removeConstraint(pc);
				delete pc;
			}
			btSoftBody* softBody = btSoftBody::upcast(obj);
			if (softBody)
			{
				mWorld->removeSoftBody(softBody);
			} 
			else
			{
				btRigidBody* body = btRigidBody::upcast(obj);
				if (body)
					mWorld->removeRigidBody(body);
				else
					mWorld->removeCollisionObject(obj);
			}
			delete obj;
		}

		// ----------------------------------------------------------------------
		mSoftBodyWorldInfo.m_sparsesdf.Reset();
	}
}


BULLETWRAPPER_API void Bullet_Step(float elapsedTime, __int32 maxSubSteps, float fixedTimestep)
{
	//mBroadPhase->resetPool(mCollisionDispatcher);


	mWorld->stepSimulation(elapsedTime, maxSubSteps, fixedTimestep);

	mSoftBodyWorldInfo.m_sparsesdf.GarbageCollect();

	//for(int i = mWorld->getNumCollisionObjects()-1; i >= 0; i--)
	//{
	//	btCollisionObject*	obj=mWorld->getCollisionObjectArray()[i];
	//	btSoftBody* softBody = btSoftBody::upcast(obj);
	//	if (softBody)
	//	{
	//		softBody->dampClusters();
	//	}
	//}
}	


BULLETWRAPPER_API __int32 Bullet_GetNumObjects()
{
	return mWorld->getNumCollisionObjects();
}



void SetUserPointer(int ObjIndex)
{	
	_int32 pt =  mWorld->getNumCollisionObjects()-1;
	btCollisionObject* co = mWorld->getCollisionObjectArray()[ObjIndex];
	co->setUserPointer((__int32 *) ObjIndex);
}




// =======================================================================================
//   BASE PLANE
// =======================================================================================

BULLETWRAPPER_API __int32 Bullet_CreateFlatPlane(float PlaneHeight)
{
	btTransform t;
	t.setIdentity();		
	btCollisionShape *shape = new btStaticPlaneShape(btVector3(0, 1, 0), PlaneHeight);
	//mCollisionShapes.push_back(shape);
	btRigidBody *body = CreateRigidBody( .0, t, shape );
	mWorld->addRigidBody(body);

	__int32 ObjIndex = mWorld->getNumCollisionObjects()-1;
	SetUserPointer(ObjIndex);
	return ObjIndex;
}


// ===========================================================================================
//   SIMPLE OBJECT ( BOX / SPHERE / CAPSULE / CONE / CYLINDER )
// ===========================================================================================

BULLETWRAPPER_API __int32 Bullet_CreateSimpleObject(D3DXVECTOR3 OffsetPosition,
													D3DXVECTOR3 ShapeSize,
													__int32 ShapeType,
													float Mass,
													float Margin, 
													bool Activated)
{

	btCollisionShape* shape = CreateSimpleShape(btVector3(1.0f, 1.0f, 1.0f), ShapeType);

	btVector3 offset_pos = ConvertVector(OffsetPosition);
	btVector3 shape_size = ConvertVector(ShapeSize);
	
	shape->setLocalScaling(shape_size);
	shape->setMargin(Margin);

	btTransform trans;
	trans.setIdentity();
	trans.setOrigin(offset_pos);

	btRigidBody* body = CreateRigidBody(Mass, trans, shape);

	if (!Activated) body->setActivationState(body->getActivationState() & !ACTIVE_TAG );

	mWorld->addRigidBody(body);

	// ------------------------------------------------ Anti tunneling by Motion Clamping
	body->setCcdMotionThreshold(1);
	body->setCcdSweptSphereRadius(0.2f);

	__int32 ObjIndex = mWorld->getNumCollisionObjects()-1;
	SetUserPointer(ObjIndex);
	return ObjIndex;
}


// =======================================================================================
//   CREATE OBJECT
// =======================================================================================


BULLETWRAPPER_API __int32 Bullet_CreateObject(float *vertices,
											  __int32 *indices,
											  __int32 numVertices,
											  __int32 NumberBytesPerVertex,
											  __int32 numFaces,
											  bool Is16BitIndices,
											  D3DXMATRIX WorldMatrix,
											  __int32 CollisionShapeType,
											  float Mass,
											  float Margin, 
											  bool Activated)
{
	btCollisionShape* shape;
	btTransform t;

	switch (CollisionShapeType)
	{
		case 0: // BOX
			{
				shape = CreateBvhTriangleMeshShape(vertices,
												   indices,
												   numVertices,
												   NumberBytesPerVertex,
												   numFaces,
												   Is16BitIndices);	

				t.setIdentity();
				btVector3 aabbMin;
				btVector3 aabbMax;
				shape->getAabb(t, aabbMin, aabbMax);
				delete shape;

				btVector3 shape_size = aabbMax - aabbMin;
				shape = CreateSimpleShape(shape_size, BOX_SHAPE_PROXYTYPE);

				shape->setLocalScaling(btVector3(1,1,1));
				shape->setMargin(Margin);

				break;
			}


		case 50: // ConvexHullShape
			shape = CreateConvexHullShape(vertices,
										  indices,
										  numVertices,
										  NumberBytesPerVertex,
										  numFaces,
										  Is16BitIndices);
			break;

		case 99: // BvhTriangleMeshShape
		default:
			shape = CreateBvhTriangleMeshShape(vertices,
											   indices,
											   numVertices,
											   NumberBytesPerVertex,
											   numFaces,
											   Is16BitIndices);
			break;
	}



	// ------------------------------------------------ extract the scale from the DirecX-Matrix
	btVector3 scale = Matrix_GetScale(WorldMatrix);
	// ------------------------------------------------ remove the scale from the DirecX-Matrix
	Matrix_SetScaleToOne(&WorldMatrix);
	// ------------------------------------------------ set the scale
	shape->setLocalScaling(scale);
	// ------------------------------------------------ set the transform ( rotation and translation )
	t = ConvertMatrix(WorldMatrix);

	shape->setMargin( Margin );

	// ------------------------------------------------
	btRigidBody *body = CreateRigidBody(Mass, t, shape);

	// ------------------------------------------------ Anti tunneling by Motion Clamping
	body->setCcdMotionThreshold(1.0f);
	body->setCcdSweptSphereRadius(0.2f);

	// ------------------------------------------------ 
	body->setContactProcessingThreshold(0.05f);
	body->setDeactivationTime(0.2f);

	//body->setDamping(0.05, 0.85);

	//body->setSleepingThresholds(0.5f, 0.5f);
	//body->setSleepingThresholds(1.6f, 2.5f);
	//body->setSleepingThresholds(3.2f, 5.0f);


	mWorld->addRigidBody(body);

	if (!Activated) body->setActivationState(body->getActivationState() & !ACTIVE_TAG );

	// ------------------------------------------------ 
	__int32 ObjIndex = mWorld->getNumCollisionObjects()-1;
	SetUserPointer(ObjIndex);
	return ObjIndex;
}


// =======================================================================================
//   METHODS 
// =======================================================================================

BULLETWRAPPER_API void Bullet_Object_ApplyCentralImpulse(__int32 ObjIndex, D3DXVECTOR3 Impulse)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	btVector3 impulse = ConvertVector(Impulse);
	if (body) body->applyCentralImpulse(impulse);
}

BULLETWRAPPER_API void Bullet_Object_ApplyDamping(__int32 ObjIndex, float timeStep)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) body->applyDamping(timeStep);
}

BULLETWRAPPER_API void Bullet_Object_ApplyTorqueImpulse(__int32 ObjIndex, D3DXVECTOR3 Torque)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	btVector3 torque = ConvertVector(Torque);
	if (body) body->applyTorqueImpulse(torque);
}

BULLETWRAPPER_API void Bullet_Object_Translate(__int32 ObjIndex, D3DXVECTOR3 Translation)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	btVector3 tran = ConvertVector(Translation);
	if (body) body->translate(tran);
}


// =======================================================================================
//   PROPERTIES SET
// =======================================================================================


// Additional damping can help avoiding lowpass jitter motion, help stability for ragdolls etc.
// Such damping is undesirable, so once the overall simulation quality 
// of the rigid body dynamics system has improved, this should become obsolete

BULLETWRAPPER_API void Bullet_Object_SetAdditionalDamping(__int32 ObjIndex, 
														  float AngularDampingThresholdSqr,
														  float AngularDampingFactor,
														  float LinearDampingThresholdSqr,
														  float LinearDampingFactor)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) 
	{
		// ----------------------------------- body propertyes changed to PUBLIC in "btRigidBody.h"
		body->m_additionalDamping = true;
		body->m_additionalAngularDampingThresholdSqr = AngularDampingThresholdSqr;
		body->m_additionalAngularDampingFactor = AngularDampingFactor;
		body->m_additionalLinearDampingThresholdSqr = LinearDampingThresholdSqr;
		body->m_additionalDampingFactor = LinearDampingFactor;
	}
}

BULLETWRAPPER_API void Bullet_Object_SetActivationState(__int32 ObjIndex, __int32 activationState)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) body->setActivationState(activationState);
	//if (body) body->forceActivationState(activationState);
	//if (body) body->setActivationState(body->getActivationState() | ACTIVE_TAG );
}

BULLETWRAPPER_API void Bullet_Object_SetDamping(__int32 ObjIndex, float LinearDamping, float AngularDamping)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) body->setDamping(LinearDamping, AngularDamping);
}

BULLETWRAPPER_API void Bullet_Object_SetFriction(__int32 ObjIndex, float Friction)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) body->setFriction(Friction);
}

BULLETWRAPPER_API void Bullet_Object_SetKinematic(__int32 ObjIndex, bool kinematic)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!body) return; 
	if (kinematic)
		body->setCollisionFlags( body->getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT);
	else
		body->setCollisionFlags( body->getCollisionFlags() & !btCollisionObject::CF_KINEMATIC_OBJECT);
}

BULLETWRAPPER_API void Bullet_Object_SetMatrix(__int32 ObjIndex, D3DXMATRIX m)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	btTransform t = ConvertMatrix(m);
	if (body) body->setWorldTransform(t);
}

BULLETWRAPPER_API void Bullet_Object_SetRestitution(__int32 ObjIndex, float Restitution)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) body->setRestitution(Restitution);
}

BULLETWRAPPER_API void Bullet_Object_SetSleepingThresholds(__int32 ObjIndex, 
														   float LinearThreshold, 
														   float AngularThreshold)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) body->setSleepingThresholds(LinearThreshold, AngularThreshold);
}








// ================================================================================================
//   PROPERTIES GET
// ================================================================================================

BULLETWRAPPER_API __int32 Bullet_Object_GetActivationState(__int32 ObjIndex)
{
	SAFE_EXIT_INT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) return body->getActivationState();
	else return 0;
}

BULLETWRAPPER_API float Bullet_Object_GetAngularDamping(__int32 ObjIndex)
{
	SAFE_EXIT_FLOAT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) return body->getAngularDamping();
	else return 0;
}

BULLETWRAPPER_API float Bullet_Object_GetFriction(__int32 ObjIndex)
{
	SAFE_EXIT_FLOAT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) return body->getFriction();
	else return 0;
}

BULLETWRAPPER_API float Bullet_Object_GetLinearDamping(__int32 ObjIndex)
{
	SAFE_EXIT_FLOAT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) return body->getLinearDamping();
	else return 0;
}

BULLETWRAPPER_API D3DXVECTOR3 Bullet_Object_GetLinearVelocity(__int32 ObjIndex)
{
	SAFE_EXIT_VEC3(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) return ConvertVector(body->getLinearVelocity());
	else return D3DXVECTOR3(0,0,0);
}

BULLETWRAPPER_API float Bullet_Object_GetMass(__int32 ObjIndex)
{
	SAFE_EXIT_FLOAT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		float mass = body->getInvMass();
		if (mass > 0) mass = 1 / mass;
		return mass;
	}
	btSoftBody* softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (softBody)
	{
		return softBody->getTotalMass( );
	}
	return 0;
}

BULLETWRAPPER_API D3DXMATRIX Bullet_Object_GetMatrix(__int32 ObjIndex)
{
	SAFE_EXIT_MATRIX(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) return ConvertMatrix(body->getWorldTransform());
	else return D3DXMATRIX(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1);
}

BULLETWRAPPER_API D3DXMATRIX Bullet_Object_GetMatrixScaled(__int32 ObjIndex)
{
	SAFE_EXIT_MATRIX(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btVector3 scale = body->getCollisionShape()->getLocalScaling();
		D3DXMATRIX s;
		D3DXMatrixScaling(&s, scale.x(), scale.y(), scale.z());

		btTransform t = body->getWorldTransform();
		D3DXMATRIX m = ConvertMatrix(t);

		return s * m;
	}
	else return D3DXMATRIX(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1);
}

BULLETWRAPPER_API float Bullet_Object_GetRestitution(__int32 ObjIndex)
{
	SAFE_EXIT_FLOAT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body) return body->getRestitution();
	else return 0.0f;
}

BULLETWRAPPER_API D3DXVECTOR3 Bullet_Object_GetPosition(__int32 ObjIndex)
{
	SAFE_EXIT_VEC3(ObjIndex)
	btVector3 pos(0,0,0);
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		pos = body->getWorldTransform().getOrigin();
		return ConvertVector(pos);
	}
	btSoftBody* softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (softBody)
	{
		for(int i = 0; i < softBody->m_nodes.size(); i++)
		{
			pos += softBody->m_nodes[i].m_x;
		}		
		pos /= (btScalar)softBody->m_nodes.size();
	}
	return ConvertVector(pos);
}



// ================================================================================================
//   SHOOT OBJECT
// ================================================================================================

BULLETWRAPPER_API void Bullet_Object_Shoot(__int32 ObjIndex, 
										   D3DXVECTOR3 Origin, 
										   D3DXVECTOR3 Destination, 
										   float ShootSpeed)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);

	btVector3 origin = ConvertVector(Origin);
	btVector3 destination = ConvertVector(Destination);

	btVector3 linVel = destination- origin;
	linVel.normalize();
	linVel *= ShootSpeed;

	btTransform startTransform;
	startTransform.setIdentity();		
	startTransform.setOrigin( origin );	

	body->getWorldTransform().setOrigin(origin);
	body->getWorldTransform().setRotation(btQuaternion(0,0,0,1));
	body->setLinearVelocity(linVel);
	body->setAngularVelocity(btVector3(0,0,-5));
	
	//body->setCcdMotionThreshold(150);
	//body->setCcdSweptSphereRadius(150.2f);
	
	body->activate();
}




// ================================================================================================
//   CONSTRAINTS
// ================================================================================================

// -------------------------------------------------------------------------------------------------
//  CONSTRAINTS - Point 
// -------------------------------------------------------------------------------------------------
BULLETWRAPPER_API void Bullet_Constraint_ObjectPoint(__int32 ObjIndex, 
													 D3DXVECTOR3 PivotPoint)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody* body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);

	btVector3 pivot = ConvertVector(PivotPoint);

	btTypedConstraint* c = new btPoint2PointConstraint(*body, pivot);

	mWorld->addConstraint(c);
}

BULLETWRAPPER_API void Bullet_Constraint_ObjectPointObject(__int32 ObjIndex_A,
														   __int32 ObjIndex_B, 
														   D3DXVECTOR3 PivotPoint_A,
														   D3DXVECTOR3 PivotPoint_B)
{
	SAFE_EXIT(ObjIndex_A)
	SAFE_EXIT(ObjIndex_B)
	btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);
	btRigidBody* body_b = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_B]);

	btVector3 pivot_a = ConvertVector(PivotPoint_A);
	btVector3 pivot_b = ConvertVector(PivotPoint_B);

	btTypedConstraint* c = new btPoint2PointConstraint(*body_a, *body_b, pivot_a, pivot_b);

	mWorld->addConstraint(c);
}



// -------------------------------------------------------------------------------------------------
//  CONSTRAINTS - Hinge (cerniera)  
// -------------------------------------------------------------------------------------------------

BULLETWRAPPER_API void Bullet_Constraint_ObjectHinge_Matrix(__int32 ObjIndex_A, 
															D3DXMATRIX Matrix_A,
															bool useReferenceFrameA)
{
	SAFE_EXIT(ObjIndex_A)
    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);

	btTransform trans_a = ConvertMatrix(Matrix_A);

	btTypedConstraint* c = new btHingeConstraint(*body_a, trans_a, useReferenceFrameA);

	mWorld->addConstraint(c);
}

BULLETWRAPPER_API void Bullet_Constraint_ObjectHinge_PivotAndAxis(__int32 ObjIndex_A, 
																  D3DXVECTOR3 Pivot_A,
																  D3DXVECTOR3 Axis_A,
																  bool useReferenceFrameA)
{
	SAFE_EXIT(ObjIndex_A)
	btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);

	btVector3 pivot_a = ConvertVector(Pivot_A);
	btVector3 axis_a = ConvertVector(Axis_A);

	btTypedConstraint* c = new btHingeConstraint(*body_a, pivot_a, axis_a, useReferenceFrameA);

	mWorld->addConstraint(c);
}

BULLETWRAPPER_API void Bullet_Constraint_ObjectHingeObject_Matrix(__int32 ObjIndex_A,
																  __int32 ObjIndex_B,
																  D3DXMATRIX Matrix_A,
																  D3DXMATRIX Matrix_B,
																  bool useReferenceFrameA)
{
	SAFE_EXIT(ObjIndex_A)
	SAFE_EXIT(ObjIndex_B)
    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);
	btRigidBody* body_b = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_B]);

	btTransform trans_a = ConvertMatrix(Matrix_A);
	btTransform trans_b = ConvertMatrix(Matrix_B);

	btTypedConstraint* c = new btHingeConstraint(*body_a, *body_b, trans_a, trans_b, useReferenceFrameA);

	mWorld->addConstraint(c);
}

BULLETWRAPPER_API void Bullet_Constraint_ObjectHingeObject_PivotAndAxis(__int32 ObjIndex_A,
																		__int32 ObjIndex_B, 
																		D3DXVECTOR3 Pivot_A,
																		D3DXVECTOR3 Pivot_B,
																		D3DXVECTOR3 Axis_A,
																		D3DXVECTOR3 Axis_B,
																		bool useReferenceFrameA)
{
	SAFE_EXIT(ObjIndex_A)
	SAFE_EXIT(ObjIndex_B)
    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);
	btRigidBody* body_b = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_B]);

	btVector3 pivot_a = ConvertVector(Pivot_A);
	btVector3 pivot_b = ConvertVector(Pivot_B);
	btVector3 axis_a = ConvertVector(Axis_A);
	btVector3 axis_b = ConvertVector(Axis_B);

	btTypedConstraint* c = new btHingeConstraint(*body_a, *body_b, pivot_a, pivot_b, axis_a, axis_b, useReferenceFrameA);

	mWorld->addConstraint(c);
}




// -------------------------------------------------------------------------------------------------
//  CONSTRAINTS - Slider (guida lineare)  
// -------------------------------------------------------------------------------------------------

BULLETWRAPPER_API void Bullet_Constraint_ObjectSlider(__int32 ObjIndex_A, 
													  D3DXMATRIX Matrix_A,
													  bool useReferenceFrameA)
{
	SAFE_EXIT(ObjIndex_A)
    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);

	btTransform trans_a = ConvertMatrix(Matrix_A);

	btTypedConstraint* c = new btSliderConstraint(*body_a, trans_a, useReferenceFrameA);

	mWorld->addConstraint(c);
}

BULLETWRAPPER_API void Bullet_Constraint_ObjectSliderObject(__int32 ObjIndex_A,
														    __int32 ObjIndex_B,
														    D3DXMATRIX Matrix_A,
														    D3DXMATRIX Matrix_B,
														    bool useReferenceFrameA)
{
	SAFE_EXIT(ObjIndex_A)
	SAFE_EXIT(ObjIndex_B)
    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);
	btRigidBody* body_b = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_B]);

	btTransform trans_a = ConvertMatrix(Matrix_A);
	btTransform trans_b = ConvertMatrix(Matrix_B);

	btTypedConstraint* c = new btSliderConstraint(*body_a, *body_b, trans_a, trans_b, useReferenceFrameA);

	mWorld->addConstraint(c);
}


// -------------------------------------------------------------------------------------------------
//  CONSTRAINTS - ConeTwist ()
// -------------------------------------------------------------------------------------------------

BULLETWRAPPER_API void Bullet_Constraint_ObjectConeTwist(__int32 ObjIndex_A, 
														 D3DXMATRIX Matrix_A)
{
	SAFE_EXIT(ObjIndex_A)
    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);

	btTransform trans_a = ConvertMatrix(Matrix_A);

	btTypedConstraint* c = new btConeTwistConstraint(*body_a, trans_a);

	mWorld->addConstraint(c);
}

BULLETWRAPPER_API void Bullet_Constraint_ObjectConeTwistObject(__int32 ObjIndex_A,
															   __int32 ObjIndex_B,
														       D3DXMATRIX Matrix_A,
														       D3DXMATRIX Matrix_B)
{
	SAFE_EXIT(ObjIndex_A)
	SAFE_EXIT(ObjIndex_B)
    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);
	btRigidBody* body_b = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_B]);

	btTransform trans_a = ConvertMatrix(Matrix_A);
	btTransform trans_b = ConvertMatrix(Matrix_B);

	btTypedConstraint* c = new btConeTwistConstraint(*body_a, *body_b, trans_a, trans_b);

	mWorld->addConstraint(c);
}



// -------------------------------------------------------------------------------------------------
//  CONSTRAINTS - Generic 6 Dof 
// -------------------------------------------------------------------------------------------------
// Lowerlimit == Upperlimit -> axis is locked.  
// Lowerlimit > Upperlimit -> axis is free  
// Lowerlimit < Upperlimit -> axis it limited in that range 
// -------------------------------------------------------------------------------------------------

//BULLETWRAPPER_API void Bullet_Constraint_Object_Generic6Dof(__int32 ObjIndex_B,
//															D3DXMATRIX Matrix_B,
//															bool useReferenceFrameB)
//{
//	SAFE_EXIT(ObjIndex_B)
//	btRigidBody* body_b = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_B]);
//
//	btTransform trans_b = ConvertMatrix(Matrix_B);
//
//	btGeneric6DofConstraint* c = new btGeneric6DofConstraint(*body_b, trans_b, useReferenceFrameB);
//
//	mWorld->addConstraint(c);
//}
//
//BULLETWRAPPER_API void Bullet_Constraint_Object_Generic6Dof_Object(__int32 ObjIndex_A,
//																   __int32 ObjIndex_B,
//																   D3DXMATRIX Matrix_A,
//																   D3DXMATRIX Matrix_B,
//																   bool useReferenceFrameA)
//{
//	SAFE_EXIT(ObjIndex_A)
//	SAFE_EXIT(ObjIndex_B)
//    btRigidBody* body_a = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_A]);
//	btRigidBody* body_b = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex_B]);
//
//	btTransform trans_a = ConvertMatrix(Matrix_A);
//	btTransform trans_b = ConvertMatrix(Matrix_B);
//
//	btGeneric6DofConstraint* c = new btGeneric6DofConstraint(*body_a, *body_b, trans_a, trans_b, useReferenceFrameA);
//	
//	btVector3 angular_low = ConvertVector(AngularLowerLimit);
//	btVector3 angular_up = ConvertVector(AngularUpperLimit);
//	btVector3 linear_low = ConvertVector(LinearLowerLimit);
//	btVector3 linear_up = ConvertVector(LinearUpperLimit);
//	c->setAngularLowerLimit(angular_low);
//	c->setAngularUpperLimit(angular_up);
//	c->setLinearLowerLimit(linear_low);
//	c->setLinearUpperLimit(linear_up);
//
//	c->setLimit(
//	c->setAngularLimits(
//	c->setLinearLimits(
//
//	mWorld->addConstraint(c);
//}







// ===========================================================================================
//   SNIPPETS
// ===========================================================================================

	//btVector3 v = mWorld->getGravity();
	//return v.y();


	//return mWorld->getNumCollisionObjects();


		//Bullet_GetCollisionObjectShape
		//Bullet_GetCollisionObjectPosition

//BULLETWRAPPER_API __int32 Bullet_GetCollisionObjectShape(__int32 ObjIndex)
//{
//	SAFE_EXIT(ObjIndex)
//	btCollisionObject* colObj = mWorld->getCollisionObjectArray()[ObjIndex];
//    btRigidBody* body = btRigidBody::upcast(colObj);
//	__int32 ObjectShape = (__int32)body->getCollisionShape();
//	return ObjectShape;
//}
//
//BULLETWRAPPER_API D3DXVECTOR3 Bullet_GetCollisionObjectPosition(__int32 ObjIndex)
//{
//	SAFE_EXIT(ObjIndex)
//	btCollisionObject* colObj = mWorld->getCollisionObjectArray()[ObjIndex];
//    btRigidBody* body = btRigidBody::upcast(colObj);
//	btVector3 CenterOfMassPosition = body->getCenterOfMassPosition();
//	return ConvertVector(CenterOfMassPosition);
//}




// ===========================================================================================
//   UNUSED - GET COLLISIONS
// ===========================================================================================


//// ===============================================================================================
////  Bullet_GetCollisions can find collision objects ID / penetration ( getDistance )
//// ===============================================================================================
////  this method has several problems 
////    - some collision are not generated 
////    - each newly created object must set ObjIndex to USER POINTER ( calling SetUserPointer )
//// ===============================================================================================
//
//BULLETWRAPPER_API __int32 Bullet_GetCollisions(float* CollisionArray, float minCollision)
//{
//	__int32 CollisionCount = 0;
//	__int32 CollisionArrayLen = mWorld->getNumCollisionObjects();
//
//	// -------------------------------------------------------------------------- empty the array
//	for (int j = 0; j < CollisionArrayLen; j++) CollisionArray[j] = 0.0f;
//
//	// -------------------------------------------------------------------------- 
//	int numManifolds = mWorld->getDispatcher()->getNumManifolds();
//	for (__int32 i=0; i < numManifolds; i++)
//	{
//		btPersistentManifold* contactManifold = mWorld->getDispatcher()->getManifoldByIndexInternal(i);
//		
//		btCollisionObject* obA = static_cast<btCollisionObject*>(contactManifold->getBody0());
//		btCollisionObject* obB = static_cast<btCollisionObject*>(contactManifold->getBody1());
//
//		//btRigidBody* rb0 = (btRigidBody*)contactManifold->getBody0();
//		//btRigidBody* rb1 = (btRigidBody*)contactManifold->getBody1();
//	
//		int numContacts = contactManifold->getNumContacts();
//		for (int j=0;j<numContacts;j++)
//		{
//			btManifoldPoint& pt = contactManifold->getContactPoint(j);
//
//			// ----------------------------------------------------------------- get collision value 
//			float coll = -pt.getDistance();
//			//float coll = -pt.m_distance1;
//			//float coll = fabs(pt.m_distance1)/4.0;
//			//float coll = fabs(pt.m_appliedImpulse) / 200.0;
//
//			if (coll > minCollision)
//			{
//				__int32 p;
//
//				// ------------------------------------------------------------- get indexes from USER POINTERS
//				p = (__int32)obA->getUserPointer();
//				if (coll > CollisionArray[p]) CollisionArray[p] = coll;
//				p = (__int32)obB->getUserPointer();
//				if (coll > CollisionArray[p]) CollisionArray[p] = coll;
//				// ------------------------------------------------------------- get indexes from ISLAND TAGS ???
//				//p = obA->getIslandTag();
//				//if (p>=0) if (coll > CollisionArray[p]) CollisionArray[p] = coll;
//				//p = obB->getIslandTag();
//				//if (p>=0) if (coll > CollisionArray[p]) CollisionArray[p] = coll;
//
//				//const btVector3& ptA = pt.getPositionWorldOnA();
//				//const btVector3& ptB = pt.getPositionWorldOnB();
//				//const btVector3& normalOnB = pt.m_normalWorldOnB;
//
//				CollisionCount += 1;
//			}
//		}
//	}
//	return CollisionCount;
//}
//
//


//// ===============================================================================================
////  Bullet_GetCollidingObject
//// ===============================================================================================
////  not working with SoftBodyes
//// ===============================================================================================

//BULLETWRAPPER_API __int32 Bullet_GetCollidingObject(__int32 ObjectIndex)
//{
//	SAFE_EXIT(ObjIndex)
//	// -------------------------------------------------------------------------- 
//	int numManifolds = mWorld->getDispatcher()->getNumManifolds();
//	for (__int32 i=0; i < numManifolds; i++)
//	{
//		btPersistentManifold* contactManifold = mWorld->getDispatcher()->getManifoldByIndexInternal(i);
//		
//		btCollisionObject* obA = static_cast<btCollisionObject*>(contactManifold->getBody0());
//		btCollisionObject* obB = static_cast<btCollisionObject*>(contactManifold->getBody1());
//
//		btRigidBody *rigidBodyA = btRigidBody::upcast(obA);
//		btRigidBody *rigidBodyB = btRigidBody::upcast(obB);
//
//		//btSoftBody *softBodyA = btSoftBody::upcast(obA);
//		//btSoftBody *softBodyB = btSoftBody::upcast(obB);
//
//		//if (softBodyA) return 1; 
//		//if (softBodyB) return 1; 
//
//		if (ObjectIndex == (__int32)rigidBodyB->getUserPointer())
//		{
//			//__int32 softIndex = (__int32)obA->getUserPointer();
//
//			//btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[softIndex]);
//
//		
//			return (__int32)rigidBodyB->getUserPointer(); 
//		}
// 	}
//	return -1;
//}





// =======================================================================================
//   VEHICLE
// ======================================================================================= 


BULLETWRAPPER_API __int32 Bullet_Vehicle_Create(float *vertices,
												__int32 *indices,
												__int32 numVertices,
												__int32 NumberBytesPerVertex,
												__int32 numFaces,
												bool Is16BitIndices,
												D3DXVECTOR3 ShapeDelta,
												D3DXVECTOR3 MassBarycenter,
												D3DXMATRIX WorldMatrix,
												float Mass,
												float Margin)
{


	// --------------------------------------------------------------------- ConvexHullShape
	btCollisionShape* ChassisShape = CreateConvexHullShape(vertices,
														   indices,
														   numVertices,
														   NumberBytesPerVertex,
														   numFaces,
														   Is16BitIndices);
	

	// ---------------------------------------------------------------------- Box shape
	//btCollisionShape* ChassisShape = new btBoxShape(btVector3(1.88f * 0.5f, 
	//															1.58f * 0.5f,
	//															4.7f * 0.5f));

	//mCollisionShapes.push_back(ChassisShape);


	
	// ----------------------------------------------------------------------- Compound Shape
	// For a btRigidBody, the world transform is == to the center of mass. 
	// The center of mass is shifted using a btCompoundShape and 
	// shifting the collision shape relative to the local rigid body origin
	// --------------------------------------------------------------------
	btCompoundShape* CompoundShape = new btCompoundShape();
	btVector3 barycenter = ConvertVector(MassBarycenter);
	btVector3 shapeDelta = ConvertVector(ShapeDelta);

	btTransform t;
	t.setIdentity();
	t.setOrigin(shapeDelta - barycenter);
	CompoundShape->addChildShape(t, ChassisShape);

	// ----------------------------------------------------------------------- Scale
	//
	// ------------------------------------------------ extract the scale from the DirecX-Matrix
	btVector3 scale = Matrix_GetScale(WorldMatrix);
	// ------------------------------------------------ remove the scale from the DirecX-Matrix
	Matrix_SetScaleToOne(&WorldMatrix);
	// ------------------------------------------------ set the scale
	CompoundShape->setLocalScaling(scale);
	// ------------------------------------------------ set the transform ( rotation and translation )
	t = ConvertMatrix(WorldMatrix);

	// ----------------------------------------------------------------------- Rigid Body
	btRigidBody* body = CreateRigidBody(Mass,
										t,
										CompoundShape);	

	body->setActivationState(DISABLE_DEACTIVATION);   	// Never deactivate the vehicle

	mWorld->addRigidBody(body);


	// ----------------------------------------------------------------------- Vehicle

	//mWorld->getBroadphase()->getOverlappingPairCache()->cleanProxyFromPairs(body->getBroadphaseHandle(), mWorld->getDispatcher());

	btRaycastVehicle::btVehicleTuning VehicleTuning;
	btVehicleRaycaster* VehicleRaycaster = new btDefaultVehicleRaycaster(mWorld);
	btRaycastVehicle* vehicle = new btRaycastVehicle(VehicleTuning, body, VehicleRaycaster);
	vehicle->setCoordinateSystem(0, 1, 2);
	mWorld->addVehicle(vehicle);

	// ----------------------------------------------------------------------- Vehicle custom props
	//  custom public vehicle properties to btRigidBody.h
	//   >> void* m_vehicle;
	// -----------------------------------------------------------------------
	body->m_vehicle = vehicle; 
	body->m_massBarycenter = barycenter;


	__int32 ObjIndex = mWorld->getNumCollisionObjects()-1;
	SetUserPointer(ObjIndex);
	return ObjIndex;
}





BULLETWRAPPER_API void Bullet_Vehicle_AddWheel(int ObjIndex,
											   float Dx,
											   float Dy,	
											   float Dz,
											   float Radius,
											   float suspensionRestLength, 
											   bool  IsFrontWheel,
											   float DirectionX,
											   float DirectionY,
											   float DirectionZ,
											   float AxleX,
											   float AxleY,
											   float AxleZ)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btRaycastVehicle* vehicle = (btRaycastVehicle*)body->m_vehicle;
		if (vehicle)
		{
			btVector3 connectionPoint(Dx, Dy, Dz);

			btRaycastVehicle::btVehicleTuning tuning;

			vehicle->addWheel(connectionPoint - body->m_massBarycenter, 
							  btVector3(DirectionX,DirectionY,DirectionZ),
							  btVector3(AxleX,AxleY,AxleZ),
							  suspensionRestLength,
							  Radius,
							  tuning,
							  IsFrontWheel);
		}
	}
}


BULLETWRAPPER_API void Bullet_Vehicle_SetWheelTuning(__int32 ObjIndex,
													 __int32 WheelIndex,
													 float SuspensionStiffness,
													 float SuspensionDamping,
													 float SuspensionCompression,
													 float WheelFriction,
													 float RollInfluence,
													 float MaxSuspensionTravelCm)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btRaycastVehicle* vehicle = (btRaycastVehicle*)body->m_vehicle;
		if (vehicle)
		{
			if (WheelIndex >= 0 && WheelIndex < vehicle->getNumWheels())
			{
				btWheelInfo& wheel = vehicle->getWheelInfo(WheelIndex);
				wheel.m_suspensionStiffness = SuspensionStiffness;
				wheel.m_wheelsDampingRelaxation = SuspensionDamping;
				wheel.m_wheelsDampingCompression = SuspensionCompression;
				wheel.m_frictionSlip = WheelFriction;
				wheel.m_rollInfluence = RollInfluence;
				wheel.m_maxSuspensionTravelCm = MaxSuspensionTravelCm;
				wheel.m_brake = 10.0f;
			}
		}
	}
}



BULLETWRAPPER_API D3DXMATRIX Bullet_Vehicle_GetMatrix(__int32 ObjIndex)
{
	SAFE_EXIT_MATRIX(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btVector3 barycenter = body->m_massBarycenter;
		D3DXMATRIX tr;
		D3DXMatrixTranslation(&tr, -barycenter.x(), -barycenter.y(), -barycenter.z());
		return tr * ConvertMatrix(body->getWorldTransform());
	}
	else 
	{
		return D3DXMATRIX(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1);
	}
}

BULLETWRAPPER_API D3DXMATRIX Bullet_Vehicle_GetMatrixScaled(__int32 ObjIndex)
{
	SAFE_EXIT_MATRIX(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btVector3 scale = body->getCollisionShape()->getLocalScaling();

		D3DXMATRIX s;
		D3DXMatrixScaling(&s, scale.x(), scale.y(), scale.z());

		btVector3 barycenter = body->m_massBarycenter;
		D3DXMATRIX tr;
		D3DXMatrixTranslation(&tr, -barycenter.x(), -barycenter.y(), -barycenter.z());
		return tr * s * ConvertMatrix(body->getWorldTransform());
	}
	else 
	{
		return D3DXMATRIX(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1);
	}
}




BULLETWRAPPER_API void Bullet_Vehicle_GetWheelMatrices(__int32 ObjIndex, D3DXMATRIX m[])
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btRaycastVehicle* vehicle = (btRaycastVehicle*)body->m_vehicle;
		if (vehicle)
		{
			for (int i = 0; i < vehicle->getNumWheels(); i++)
			{
				// ------------- synchronize wheels with (not interpolated) chassis transform
				//vehicle->updateWheelTransform(i, false);				

				m[i] = ConvertMatrix(vehicle->getWheelInfo(i).m_worldTransform);
			}
		}
	}
}

BULLETWRAPPER_API void Bullet_Vehicle_SetEngineForce(__int32 ObjIndex, __int32 WheelIndex, float Force)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btRaycastVehicle* vehicle = (btRaycastVehicle*)body->m_vehicle;
		if (vehicle)
		{
			if (WheelIndex >= 0 && WheelIndex < vehicle->getNumWheels())
			{
				vehicle->applyEngineForce(Force, WheelIndex);
				//vehicle->getRigidBody()->setActivationState(ACTIVE_TAG);
			}
		}
	}
}

BULLETWRAPPER_API void Bullet_Vehicle_SetBrakingForce(__int32 ObjIndex, __int32 WheelIndex, float Force)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btRaycastVehicle* vehicle = (btRaycastVehicle*)body->m_vehicle;
		if (vehicle)
		{
			if (WheelIndex >= 0 && WheelIndex < vehicle->getNumWheels())
			{
				vehicle->setBrake(Force, WheelIndex);
				//vehicle->getRigidBody()->setActivationState(ACTIVE_TAG);
			}
		}
	}
}

BULLETWRAPPER_API void Bullet_Vehicle_SetSteeringValue(__int32 ObjIndex, __int32 WheelIndex, float SteeringValue)
{
	SAFE_EXIT(ObjIndex)
	btRigidBody *body = btRigidBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (body)
	{
		btRaycastVehicle* vehicle = (btRaycastVehicle*)body->m_vehicle;
		if (vehicle)
		{
			if (WheelIndex >= 0 && WheelIndex < vehicle->getNumWheels())
			{
				vehicle->setSteeringValue(SteeringValue, WheelIndex);
				//vehicle->getRigidBody()->setActivationState(ACTIVE_TAG);
			}
		}
	}
}









// =======================================================================================
//   SOFT BODY - TUNING      TUNE-PARAMS / TUNE-VOLUME-PARAMS / USE-CLUSTERS
// =======================================================================================

BULLETWRAPPER_API void Bullet_SoftObj_TuneParams(__int32 ObjIndex,
												 float LinearStiffness,
                                                 float AngularStiffness,
                                                 float Damping,
												 float DynamicFriction,
												 float RigidContactHardness,
												 float KineticContactHardness,
												 float SoftContactHardness,
												 float AnchorHardness)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return;
	softBody->m_materials[0]->m_kLST	= LinearStiffness;			// [0,1]	0.3 to 0.1
	softBody->m_materials[0]->m_kAST	= AngularStiffness;			// [0,1]	0.3 to 0.1

	softBody->m_cfg.kDP					= Damping;					// [0,1]	### 0.02
	softBody->m_cfg.kDF					= DynamicFriction;			// [0,1]	### 0.02
	softBody->m_cfg.kCHR				= RigidContactHardness;		// [0,1]
	softBody->m_cfg.kKHR				= KineticContactHardness;	// [0,1]
	softBody->m_cfg.kSHR				= SoftContactHardness;		// [0,1]
	softBody->m_cfg.kAHR				= AnchorHardness;			// [0,1]
}

BULLETWRAPPER_API void Bullet_SoftObj_SetDamping(__int32 ObjIndex,
												 float Damping)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return;

	softBody->m_cfg.kDP					= Damping;					// [0,1]	### 0.02
}

BULLETWRAPPER_API float Bullet_SoftObj_GetDamping(__int32 ObjIndex)
{
	SAFE_EXIT_FLOAT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return 0;

	return softBody->m_cfg.kDP;
}


BULLETWRAPPER_API void Bullet_SoftObj_TuneAeroParams(__int32 ObjIndex,
													 D3DXVECTOR3 WindVelocity,
													 float AeroLift,
													 float AeroDrag)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return;

	softBody->setWindVelocity(ConvertVector(WindVelocity));

	softBody->m_cfg.kLF					= AeroLift;				// Lift coefficient [0,+inf]		### 0.1
	softBody->m_cfg.kDG					= AeroDrag;				// Drag coefficient [0,+inf]		### 0.01
	softBody->m_cfg.aeromodel			= btSoftBody::eAeroModel::V_TwoSided;
}




BULLETWRAPPER_API void Bullet_SoftObj_TuneVolumeParams(__int32 ObjIndex,
														 float VolumeStiffness,
														 float PressureCoeff,
														 float VolumeConversationCoeff,
														 float PoseMatchingCoeff,
														 float MaximumVolumeRatioForPose)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return;

	softBody->m_materials[0]->m_kVST	= VolumeStiffness;			// Volume stiffness coefficient [0,1]		0.3 
	softBody->m_cfg.kPR					= PressureCoeff;			// Pressure coefficient [-inf,+inf]			0.1
	softBody->m_cfg.kVC					= VolumeConversationCoeff;	// Volume conversation coefficient [0,+inf]	0.0
	softBody->m_cfg.kMT					= PoseMatchingCoeff;		// Pose matching coefficient [0,1]			0.0
	softBody->m_cfg.maxvolume			= MaximumVolumeRatioForPose;// Maximum volume ratio for pose			1.0 
	
}


BULLETWRAPPER_API void Bullet_SoftObj_TuneSolverParams( __int32 ObjIndex,
														  float VelocitiesCorrectionFactor,
														__int32 VelocitiesSolverIterations,
														__int32 PositionsSolverIterations,
														__int32 DriftSolverIterations,
														__int32 ClusterSolverIterations,
														  float Timescale)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return;
	
	softBody->m_cfg.kVCF		= VelocitiesCorrectionFactor;	// (default 1) 1.0f (Baumgarte)
	softBody->m_cfg.viterations	= VelocitiesSolverIterations;	// (default 0) 0
	softBody->m_cfg.piterations = PositionsSolverIterations;	// (default 1) ### 5 (increase for no explosions)
	softBody->m_cfg.diterations	= DriftSolverIterations;		// (default 0) 0
	softBody->m_cfg.citerations	= ClusterSolverIterations;		// (default 4) 4
	softBody->m_cfg.timescale	= 1.0f;//Timescale;				// (default 1) 1.0f

	//softBody->setSolver(btSoftBody::eSolverPresets::Velocities);
}




BULLETWRAPPER_API void Bullet_SoftObj_UseClusters(__int32 ObjIndex,
												  __int32 ClusterCount,
												  float SoftVsRigidHardness,
												  float SoftVsRigidImpulseSplit,
												  float SoftVsSoftHardness,
												  float SoftVsSoftImpulseSplit, 
												  float SoftVsKineticHardness,
												  float SoftVsKineticImpulseSplit)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return;

	// ------------------------------------------------------------------------------------- CLUSTERS
	//  zero = cluster for each tetrahedron or triangle     ( 11x11 = 200 clusters )
	//  other values ( 1000 .. 10000 )					    ( 11x11 = 40 clusters )
	//  value 70											( 11x11 = 53 clusters )
	// ----------------------------------------------------------------------------------------------
	softBody->generateClusters(ClusterCount);

	__int32 n = softBody->clusterCount();
	for (int i = 0; i < n; i++)
	{
		softBody->m_clusters[i]->m_adamping = 0.1f;
		softBody->m_clusters[i]->m_ndamping = 0.1f;
		softBody->m_clusters[i]->m_ldamping = 0.1f;
		//softBody->m_clusters[i]->m_maxSelfCollisionImpulse = 0.1;
		//softBody->m_clusters[i]->m_selfCollisionImpulseFactor = 0.1;
	}

	softBody->m_cfg.collisions = btSoftBody::fCollision::CL_RS;		// rigid body - soft body 
	softBody->m_cfg.collisions += btSoftBody::fCollision::CL_SS;	// soft body - soft body
	softBody->m_cfg.collisions += btSoftBody::fCollision::CL_SELF;	// self collisions

	softBody->m_cfg.kSRHR_CL		= SoftVsRigidHardness;		// [0,1] 	### 1.0(helps to stop penetrations)
	softBody->m_cfg.kSR_SPLT_CL		= SoftVsRigidImpulseSplit;  // [0,1]	### 0.1(helps to stop penetrations)

	softBody->m_cfg.kSSHR_CL		= SoftVsSoftHardness;		// [0,1]	### 0.1(no explosions)
	softBody->m_cfg.kSS_SPLT_CL		= SoftVsSoftImpulseSplit;	// [0,1]	0.1	

	softBody->m_cfg.kSKHR_CL		= SoftVsKineticHardness;	// [0,1]	0.1
	softBody->m_cfg.kSK_SPLT_CL		= SoftVsKineticImpulseSplit;// [0,1]	0.1	
}





// =======================================================================================
//   SOFT BODY - CREATION
// =======================================================================================


BULLETWRAPPER_API __int32 Bullet_SoftObj_CreateSoftObject(float *vertices,
														  __int32 *indices,
														  __int32 numVertices,
														  __int32 NumberBytesPerVertex,
														  __int32 numFaces,
														  bool Is16BitIndices,
														  D3DXMATRIX WorldMatrix,
														  float Mass,
														  float Margin,
														  __int32 BendingConstraintsDist)
{

	// ------------------------------------------------------------- CREATE trimeshshape and SoftBody
	TriMeshData* trimeshshape = CreateTriMeshData(vertices,
												  indices,
												  numVertices,
												  NumberBytesPerVertex,
												  numFaces,
												  Is16BitIndices);

	btSoftBody*	softBody = btSoftBodyHelpers::CreateFromTriMesh(mSoftBodyWorldInfo,
																trimeshshape->vertices,
																trimeshshape->indices,
																numFaces,
																true);

	// ---------------------------------------------------------- delete trimeshshape
	//	TODO - delete crashes
	//delete[] trimeshshape->indices;
	//delete[] trimeshshape->vertices;
	//delete trimeshshape;


	// ------------------------------------------------------------------------------------- scale
	softBody->scale(Matrix_GetScale(WorldMatrix));
	// ------------------------------------------------------------------------------------- rotate
	btTransform t = ConvertMatrix(WorldMatrix);
	softBody->rotate(t.getRotation());
	// ------------------------------------------------------------------------------------- translate
	softBody->translate(t.getOrigin());


	// ------------------------------------------------------------------------------------- add to world
	mWorld->addSoftBody(softBody);
	// ------------------------------------------------------------------------------------- get ObjIndex
	__int32 ObjIndex = mWorld->getNumCollisionObjects()-1;
	SetUserPointer(ObjIndex);

	// ------------------------------------------------------------------------------------- mass and margin
	softBody->setTotalMass(Mass, true);
	softBody->getCollisionShape()->setMargin(Margin);	//		### 0.1 .. 0.2


	// -------------------------------------------  Anti tunneling by Motion Clamping ( not working on SoftBody )
	//softBody->setCcdMotionThreshold(1);
	//softBody->setCcdSweptSphereRadius(0.2f);


	// ------------------------------------------------------------------------------------- bending constraints
	if (BendingConstraintsDist > 0)
	{
		softBody->generateBendingConstraints(BendingConstraintsDist);	//	### 2..4
		softBody->randomizeConstraints();
	}

	// ------------------------------------------------------------------------------------- COLLISIONS
	softBody->m_cfg.collisions = btSoftBody::fCollision::SDF_RS; // SDF based rigid vs soft
	softBody->m_cfg.collisions += btSoftBody::fCollision::VF_SS; // Vertex vs face soft vs soft handling


	// ------------------------------------------------------------------------------------- pose
	softBody->setPose(true, true);
	//softBody->setTotalDensity(0.001f);
	//softBody->setVolumeDensity(0.001f);
	//softBody->setVolumeMass(0.001f);


	// ------------------------------------------------------------------------------------- DEFAULT PARAMS
	Bullet_SoftObj_TuneParams(ObjIndex, 0.3f, 0.3f, 0.01f, 0.02f, 0.5f, 0.3f, 0.3f, 0.1f);

	// ------------------------------------------------------------------------------------- DEFAULT SOLVER PARAMS
	Bullet_SoftObj_TuneSolverParams(ObjIndex, 1.0, 0, 2, 0, 4, 1.0);
	

	// ------------------------------------------------------------------------------------- return
	return ObjIndex;
}





BULLETWRAPPER_API __int32 Bullet_SoftObj_CreateCloth(
													 float Size_x,
													 float Size_z, 
													 __int32 NodeCount_x,
													 __int32 NodeCount_z,
													 float Pos_x,
													 float Pos_y,
													 float Pos_z,
													 float Yaw,
													 float Pitch,
													 float Roll,
													 __int32 FixedSides,
													 float TotalMass,
													 float Margin,
													 __int32 BendingConstraintsDist)
{

	Size_x /= 2;
	Size_z /= 2;

	btSoftBody*	softBody = btSoftBodyHelpers::CreatePatch(mSoftBodyWorldInfo,
														  btVector3(-Size_x,0,-Size_z),
														  btVector3(+Size_x,0,-Size_z),
														  btVector3(-Size_x,0,+Size_z),
														  btVector3(+Size_x,0,+Size_z),
														  NodeCount_x,
														  NodeCount_z,
														  FixedSides,
														  false); 
	// use false and BendingConstraints = 2
	//  or true and no GenerateBendingConstraints

	// ------------------------------------------------------------------------------------- rotate
	D3DXMATRIX m;
	D3DXMatrixRotationYawPitchRoll(&m, Yaw, Pitch, Roll);
	btTransform t = ConvertMatrix(m);
	softBody->rotate(t.getRotation());

	// ------------------------------------------------------------------------------------- translate
	softBody->translate(btVector3(Pos_x, Pos_y, Pos_z));

	// ------------------------------------------------------------------------------------- add to world
	mWorld->addSoftBody(softBody);
	// ------------------------------------------------------------------------------------- get ObjIndex
	__int32 ObjIndex = mWorld->getNumCollisionObjects()-1;
	SetUserPointer(ObjIndex);


	// ------------------------------------------------------------------------------------- mass and margin
	softBody->setTotalMass(TotalMass);
	softBody->getCollisionShape()->setMargin(Margin);

	// ------------------------------------------------------------------------------------- bending constraints
	if (BendingConstraintsDist > 0)
	{
		softBody->generateBendingConstraints(BendingConstraintsDist); // ### 2..4
		softBody->randomizeConstraints();
	}

	// ------------------------------------------------------------------------------------- COLLISIONS
	softBody->m_cfg.collisions = btSoftBody::fCollision::SDF_RS; // SDF based rigid vs soft
	softBody->m_cfg.collisions += btSoftBody::fCollision::VF_SS; // Vertex vs face soft vs soft handling




	// ------------------------------------------- Anti tunneling by Motion Clamping ( not working on cloths )
	//softBody->setCcdMotionThreshold(.1);
	// CCD works on an embedded sphere of radius, make sure this radius is embedded inside the convex objects, preferably smaller: for an object of dimentions 1 meter, try 
	//softBody->setCcdSweptSphereRadius(0.2f);


	// ------------------------------------------------------------------------------------- DEFAULT PARAMS
	Bullet_SoftObj_TuneParams(ObjIndex, 0.3f, 0.3f, 0.01f, 0.02f, 0.5f, 0.3f, 0.3f, 0.1f);

	// ------------------------------------------------------------------------------------- DEFAULT SOLVER PARAMS
	Bullet_SoftObj_TuneSolverParams(ObjIndex, 1.0, 0, 6, 0, 4, 1.0);


	// ------------------------------------------------------------------------------------- return
	return ObjIndex;
}






//btSoftBody * btSoftBodyHelpers::CreateFromConvexHull  ( btSoftBodyWorldInfo &  worldInfo,  
//													  const btVector3 *  vertices,  
//													  int  nvertices,  
//													  bool  randomizeConstraints = true   
//													 ) 



	//softBody->appendAnchor(0, rigidBody);
	//softBody->appendAnchor(r-1, rigidBody);




BULLETWRAPPER_API __int32 Bullet_SoftObj_CreateRope(D3DXVECTOR3 From,
													D3DXVECTOR3 To,
													__int32 SegmentCount,
													__int32 FixedSides,
													float TotalMass,
													float Margin,
													__int32 BendingConstraintsDist)
{	
	// ----------------------------------------------------------------------
	//  "CreateRope" generates a number of nodes = SegmentCount + 2
	//                             so here SegmentCount must be reduced by 1
	// ----------------------------------------------------------------------
	btSoftBody*	softBody = btSoftBodyHelpers::CreateRope(mSoftBodyWorldInfo,
														 ConvertVector(From),
														 ConvertVector(To),
														 SegmentCount - 1,
														 FixedSides);

	// ----------------------------------------------------------------------------- add to world
	mWorld->addSoftBody(softBody);
	// ----------------------------------------------------------------------------- obj index
	__int32 ObjIndex = mWorld->getNumCollisionObjects()-1;
	SetUserPointer(ObjIndex);

	// ----------------------------------------------------------------------------- mass and margin
	softBody->setTotalMass(TotalMass);
	softBody->getCollisionShape()->setMargin(Margin);


	// -----------------------------------------------------------------------------
	if (BendingConstraintsDist > 0)
	{
		softBody->generateBendingConstraints(BendingConstraintsDist); //	### 2..4
		softBody->randomizeConstraints();
	}

	// ------------------------------------------------------------------------------------- COLLISIONS
	softBody->m_cfg.collisions = btSoftBody::fCollision::SDF_RS;	// SDF based rigid vs soft
	softBody->m_cfg.collisions += btSoftBody::fCollision::VF_SS;	// Vertex vs face soft vs soft handling



	// ------------------------------------------------------------------------------------- DEFAULT PARAMS
	Bullet_SoftObj_TuneParams(ObjIndex, 0.9f, 0.9f, 0.001f, 0.001f, 0.5f, 0.9f, 0.5f, 0.9f);

	// ------------------------------------------------------------------------------------- DEFAULT SOLVER PARAMS
	Bullet_SoftObj_TuneSolverParams(ObjIndex, 1.0, 0, 8, 0, 4, 1.0);


	return ObjIndex;
}








// =======================================================================================
//   SOFT BODY - GET and SET
// =======================================================================================

BULLETWRAPPER_API __int32 Bullet_SoftObj_GetNodesCount(__int32 ObjIndex)
{
	SAFE_EXIT_INT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (softBody) return softBody->m_nodes.size();
	else return 0;
}

BULLETWRAPPER_API void Bullet_SoftObj_GetNodes(__int32 ObjIndex, D3DXVECTOR3* nodes, __int32 NodesCount)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return; 

	if (NodesCount <= 0) NodesCount = softBody->m_nodes.size();
	for(int i = 0; i < NodesCount; i++)
	{
		nodes[i].x = softBody->m_nodes[i].m_x.getX();
		nodes[i].y = softBody->m_nodes[i].m_x.getY();
		nodes[i].z = softBody->m_nodes[i].m_x.getZ();
	}
}

BULLETWRAPPER_API __int32 Bullet_SoftObj_GetFacesCount(__int32 ObjIndex)
{
	SAFE_EXIT_INT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (softBody) return softBody->m_faces.size();
	else return 0;
}


BULLETWRAPPER_API void Bullet_SoftObj_GetFaces(__int32 ObjIndex, __int32* faces)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (!softBody) return; 

	btSoftBody::Node* nbase = &softBody->m_nodes[0];
	__int32 facesCount = softBody->m_faces.size();

	for(int i = 0; i < facesCount; i++)
	{
		faces[i*3+0] = __int32(softBody->m_faces[i].m_n[0]-nbase);
		faces[i*3+1] = __int32(softBody->m_faces[i].m_n[1]-nbase);
		faces[i*3+2] = __int32(softBody->m_faces[i].m_n[2]-nbase);
	}
}


BULLETWRAPPER_API void Bullet_SoftObj_ApplyForce(__int32 ObjIndex, D3DXVECTOR3 Force, __int32 Node)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (softBody) softBody->addForce(ConvertVector(Force), Node);
}




// =======================================================================================
//   SOFT BODY - CUT
// =======================================================================================


struct	ImplicitSphere : btSoftBody::ImplicitFn
{
	btVector3	center;
	btScalar	sqradius;
	ImplicitSphere() {}
	ImplicitSphere(const btVector3& c,btScalar r) : center(c),sqradius(r*r) {}
	btScalar	Eval(const btVector3& x)
	{
		return((x-center).length2()-sqradius);
	}
};

BULLETWRAPPER_API void Bullet_SoftObj_Cut(__int32 ObjIndex, 
										  D3DXVECTOR3 Position,
										  float Radius,
										  float Accuracy)
{
	SAFE_EXIT(ObjIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[ObjIndex]);
	if (softBody) 
	{
		ImplicitSphere	isphere(ConvertVector(Position), Radius);
		softBody->refine(&isphere, Accuracy, true);
	}
}




// =======================================================================================
//   SOFT BODY - LINKS
// =======================================================================================


BULLETWRAPPER_API void Bullet_SoftObj_LinkToRigid(__int32 SoftBodyIndex, 
												  __int32 Node,
												  __int32 RigidBodyIndex, 
												  bool DisableCollisionBetweenLinkedBodies)
{
	SAFE_EXIT(SoftBodyIndex)
	SAFE_EXIT(RigidBodyIndex)
	btSoftBody *softBody = btSoftBody::upcast(mWorld->getCollisionObjectArray()[SoftBodyIndex]);
	btRigidBody *rigidBody = btRigidBody::upcast(mWorld->getCollisionObjectArray()[RigidBodyIndex]);
	if (!softBody || !rigidBody) return;  
	
	// ----------------------------------------------- use negative numbers to specify the end side
	if (Node < 0) Node += softBody->m_nodes.size();
	// ----------------------------------------------- limit to valid nodes
	Node = max(Node, 0);
	Node = min(Node, softBody->m_nodes.size() - 1);

	softBody->appendAnchor(Node, rigidBody ,DisableCollisionBetweenLinkedBodies);
}








//// ???
//BULLETWRAPPER_API void Bullet_SoftObj_AppendLinearJoint(__int32 SoftBodyIndex1, 
//													    __int32 SoftBodyIndex2, 
//													    __int32 Node1,
//													    __int32 Node2)
//{
//	SAFE_EXIT(ObjIndex)
//	SAFE_EXIT(ObjIndex)
//	btSoftBody *softBody1 = btSoftBody::upcast(mWorld->getCollisionObjectArray()[SoftBodyIndex1]);
//	btSoftBody *softBody2 = btSoftBody::upcast(mWorld->getCollisionObjectArray()[SoftBodyIndex2]);
//	if (softBody1 && softBody2) 
//		softBody1->appendLinearJoint(const LJoint::Specs& specs,btSoftBody2);
//}
//
//
//// Hinge
//BULLETWRAPPER_API void Bullet_SoftObj_AppendAngularJoint(__int32 SoftBodyIndex1, 
//													    __int32 SoftBodyIndex2, 
//													    __int32 Node1,
//													    __int32 Node2)
//{
//	SAFE_EXIT(ObjIndex)
//	SAFE_EXIT(ObjIndex)
//	btSoftBody *softBody1 = btSoftBody::upcast(mWorld->getCollisionObjectArray()[SoftBodyIndex1]);
//	btSoftBody *softBody2 = btSoftBody::upcast(mWorld->getCollisionObjectArray()[SoftBodyIndex2]);
//	if (softBody1 && softBody2) 
//		SoftBodyIndex1->appendAngularJoint(const AJoint::Specs& specs,Body body=Body());
//}



