
#include "stdafx.h"

#include "DXUT_BW.h"

// Bullet includes
#include <btBulletDynamicsCommon.h>
#include <LinearMath/btQuickprof.h>
#include <BulletCollision/CollisionDispatch/btGhostObject.h>
#include <BulletCollision/CollisionShapes/btHeightfieldTerrainShape.h>
#include <BulletDynamics/Character/btKinematicCharacterController.h>

struct TriMeshData
{
	btScalar	*vertices;
	int			*indices;	
	btTriangleIndexVertexArray	*indexVertexArrays;
	TriMeshData()
	{
		vertices = NULL;
		indices = NULL;
		indexVertexArrays = NULL;
	}
	~TriMeshData()
	{
		if( vertices ) delete [] vertices;
		if( indices ) delete [] indices;
		if( indexVertexArrays ) delete indexVertexArrays;
	}
};

D3DXMATRIX ConvertMatrix( btTransform &trn );
btTransform ConvertMatrix( D3DXMATRIX &mat );
btVector3 Matrix_GetScale(D3DXMATRIX &m);
void Matrix_SetScaleToOne(D3DXMATRIX* m);

D3DXVECTOR3 ConvertVector( btVector3 &vec );
D3DXVECTOR3 ConvertVector( const btVector3 &vec );
btVector3 ConvertVector( D3DXVECTOR3 &vec );
btVector3 ConvertVector( const D3DXVECTOR3 *vec );

const btVector3 rotate( const btQuaternion& quat, const btVector3 & vec );

TriMeshData* CreateTriMeshData(float *pVertices,
						      __int32 *pIndices,
						      __int32 numVertices,
						      __int32 NumberBytesPerVertex,
						      __int32 numFaces,
						      bool Is16BitIndices);

btRigidBody* CreateRigidBody(float mass, 
							 const btTransform& startTransform, 
							 btCollisionShape* shape);


btCollisionShape* CreateSimpleShape(const btVector3& ShapeSize, int shapeType);

btRigidBody* CreateSimpleShapeBody(const btVector3& offsetPosition, 
								   const btVector3& shapeSize, 
								   int shapeType,
								   float Margin,
								   float Mass);

btBvhTriangleMeshShape *CreateBvhTriangleMeshShape(float *pVertices,
												   __int32 *pIndices,
												   __int32 numVertices,
												   __int32 NumberBytesPerVertex,
												   __int32 numFaces,
												   bool Is16BitIndices);

btConvexHullShape *CreateConvexHullShape(float *pVertices,
									     __int32 *pIndices,
									     __int32 numVertices,
									     __int32 NumberBytesPerVertex,
									     __int32 numFaces,
									     bool Is16BitIndices);